#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""APNsClientStream class definition."""

from __future__ import annotations

import asyncio
import logging
import socket
from asyncio import StreamReader, StreamWriter
from logging import getLogger
from typing import TYPE_CHECKING, Final, Self

from ..._util.crypto import randint
from ..streams._apns_stream import ConnectionType, _APNsStream

if TYPE_CHECKING:
    from types import TracebackType

    from ..protocol.packet import APNsCommand

COURIER_HOSTNAME_TEMPLATE: Final = "{0}-courier.push.apple.com"

logger = getLogger(__name__)


class APNsClientStream(_APNsStream):
    """Stream class for client connections to the APNs courier."""

    def __init__(self: Self, host: str | None = None, port: int = 5223) -> None:
        """Initialize a new APNs client stream."""
        if host is not None:
            addr_info = socket.getaddrinfo(host, port)
            host, port, *_ = addr_info[0][4]

        super().__init__(ConnectionType.CLIENT, host, port)

        self._stream_writer: StreamWriter | None = None
        self._stream_reader: StreamReader | None = None
        self._logger = getLogger(f"{__name__}.{self.__class__.__name__}")

    async def __aenter__(self: Self) -> Self:
        """Async context manager enter method."""
        await self.connect()
        return self

    async def __aexit__(
        self: Self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Async context manager exit method."""
        await self.close()

    def log(
        self: Self,
        level: int,
        msg: str,
        *args: str,
        client_address: tuple[str, int] | None = None,
        **kwargs: ...,
    ) -> None:
        """Log a message with the client address if available."""
        addr_repr = f"[{client_address[0]}:{client_address[1]}] " if client_address else ""
        self._logger.log(level, f"Client :: {addr_repr}{msg}", *args, **kwargs, stacklevel=2)

    @staticmethod
    def fetch_random_courier_address() -> str:
        """Generate a random APNs courier address."""
        return COURIER_HOSTNAME_TEMPLATE.format(randint(1, 50))

    async def connect(self: Self) -> None:
        """Connect to the APNs courier and perform the TLS handshake."""
        if self.host is None:
            self.host = self.fetch_random_courier_address()

        self.log(logging.DEBUG, f"Connecting to APNs courier at {self.host}:{self.port}...")

        self._stream_reader, self._stream_writer = await asyncio.open_connection(
            self.host,
            self.port,
            ssl=self._create_ssl_context(ConnectionType.CLIENT),
        )

        self.log(logging.DEBUG, "TLS handshake successful and connection established.")

    async def close(self: Self) -> None:
        """Close the connection to the APNs courier."""
        self.log(logging.DEBUG, "Closing connection to APNs courier...")
        self._stream_writer.close()
        await self._stream_writer.wait_closed()

    async def read(self: Self) -> APNsCommand:
        """Read a command from the stream."""
        result = await super()._read_command(self._stream_reader)
        self.log(logging.DEBUG, f"Received {result.__class__.__name__} (0x{result.command_id:02X}).")
        return result

    async def send(self: Self, command: APNsCommand) -> None:
        """Send a command to the stream."""
        self.log(logging.DEBUG, f"Sending {command.__class__.__name__} (0x{command.command_id:02X}).")
        await super()._send_command(self._stream_writer, command)



#  Copyright (C) 2024  Cypheriel
"""Main entrypoint for the PommeKit CLI."""

import logging
from datetime import datetime
from sys import stderr
from typing import Annotated

import typer
from rich.console import Console
from rich.logging import RichHandler

from ._cli import device
from ._cli.util.app_dirs import USER_LOG_DIR
from ._cli.util.rich_console import console

app = typer.Typer()
app.add_typer(device.app)
app.add_typer(device.app, name=device.__ALIAS__, hidden=True)

logger = logging.getLogger(__name__)


def _setup_logging(level: int) -> None:
    logging_file = USER_LOG_DIR / datetime.now().astimezone().strftime("%Y-%m-%d_%H-%M-%S.log")
    logging_file.parent.mkdir(parents=True, exist_ok=True)

    logging.basicConfig(
        level=level,
        format="%(message)s",
        handlers=[
            RichHandler(
                console=Console(file=stderr),
            ),
            logging.FileHandler(logging_file),
        ],
    )


@app.callback(no_args_is_help=True)
def main(
    *,
    verbose: Annotated[
        bool,
        typer.Option("--verbose", "-v", help="Enable verbose logging."),
    ] = False,
    quiet: Annotated[
        bool,
        typer.Option("--quiet", "-q", help="Enable quiet logging."),
    ] = False,
    silent: Annotated[
        bool,
        typer.Option("--silent", help="Completely disable logging."),
    ] = False,
    no_color: Annotated[
        bool,
        typer.Option("--no-color", envvar="NO_COLOR", help="Disable color output."),
    ] = False,
) -> None:
    """PommeKit - Python library with various tools for interacting with Apple services and APIs."""
    if (verbose, quiet, silent).count(True) > 1:
        typer.echo("Can only enable one of --verbose, --quiet, or --silent.", err=True)
        raise typer.Abort

    if not any((verbose, quiet, silent)):
        _setup_logging(logging.INFO)
        logger.debug("Logging initialized at INFO level.")

    if verbose:
        _setup_logging(logging.DEBUG)
        logger.debug("Logging initialized at DEBUG level.")

    if quiet:
        _setup_logging(logging.CRITICAL)
        logger.debug("Logging initialized at CRITICAL level.")

    if silent:
        logging.disable(logging.CRITICAL)

    if no_color:
        console.no_color = True


__entrypoint__ = app
if __name__ == "__main__":
    __entrypoint__()



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""Module containing the base class responsible for serializing and deserializing APNs command packets."""

from __future__ import annotations

from collections.abc import Collection
from dataclasses import MISSING, dataclass, field, fields, is_dataclass
from io import BytesIO
from logging import getLogger
from typing import Annotated, ClassVar, Final, Self, TypeVar, get_origin

from ..protocol.transformers import NO_OP_TRANSFORMER, DataTransformer

_T = TypeVar("_T")

# Packet protocol definitions
# Format: [COMMAND_ID][COMMAND_LENGTH][ITEM_ID][ITEM_LENGTH][ITEM_DATA][ITEM_ID][ITEM_LENGTH][ITEM_DATA]...
COMMAND_ID_BYTES: Final = 1
COMMAND_LENGTH_BYTES: Final = 4
ITEM_ID_BYTES: Final = 1
ITEM_LENGTH_BYTES: Final = 2

logger = getLogger(__name__)


def _serialize_item(metadata: Item, value: _T) -> bytes:
    """Serialize an item using the provided metadata and value."""
    serialized_value = metadata.transformer.serialize(value)
    return (
        metadata.item_id.to_bytes(ITEM_ID_BYTES) + len(serialized_value).to_bytes(ITEM_LENGTH_BYTES) + serialized_value
    )


def _get_default_field_values(cls: type) -> dict[str, ...]:
    """Get the default values for all fields in the command class."""
    if is_dataclass(cls) is False:
        msg = "The provided class is not a dataclass!"
        raise ValueError(msg)

    # noinspection PyDataclass
    return {
        command_field.name: (
            command_field.default
            if command_field.default is not MISSING
            else command_field.default_factory()
            if callable(command_field.default_factory)
            else None
        )
        for command_field in fields(cls)
    }


@dataclass
class Item:
    """Simple dataclass to encapsulate metadata about a command field."""

    item_id: int
    transformer: DataTransformer = NO_OP_TRANSFORMER


@dataclass(kw_only=True)
class APNsCommand:
    """Base class for APNs commands."""

    command_id: ClassVar[int]

    unknown_items: list[tuple[int, bytes]] = field(default_factory=list)
    """List of items that could not be deserialized. Each item is a tuple of ``(item_id, item_data)``."""

    def __bytes__(self: Self) -> bytes:
        """Serialize the command to bytes."""
        if not hasattr(self, "command_id"):
            msg = "Missing command ID in command. Please register the command first!"
            raise ValueError(msg)

        # Sort fields by item ID
        sorted_fields = sorted(
            fields(self),
            key=lambda f: f.type.__metadata__[0].item_id if get_origin(f.type) is Annotated else -1,
        )

        item_data = b""

        for command_field in sorted_fields:
            if get_origin(command_field.type) is not Annotated:
                continue

            metadata = command_field.type.__metadata__
            if len(metadata) != 1 or not isinstance(metadata[0], Item):
                continue

            metadata = metadata[0]

            value = getattr(self, command_field.name)

            if value is None:
                continue

            if isinstance(value, Collection) and not isinstance(value, (str, bytes)):
                for item in value:
                    item_data += _serialize_item(metadata, item)

            else:
                item_data += _serialize_item(metadata, value)

        return self.command_id.to_bytes(COMMAND_ID_BYTES) + len(item_data).to_bytes(COMMAND_LENGTH_BYTES) + item_data

    @classmethod
    def from_bytes(cls: type[Self], data: bytes, *, includes_header: bool = False) -> type[Self]:
        """
        Deserialize a command from bytes.

        :param data: The bytes to deserialize.
        :param includes_header: Whether the data includes the command ID and length.
        """
        stream = BytesIO(data)

        if includes_header:
            command_id = int.from_bytes(stream.read(COMMAND_ID_BYTES))
            if command_id != cls.command_id:
                msg = "The provided data does not match the command ID of this class!"
                raise ValueError(msg)

            stream.read(COMMAND_LENGTH_BYTES)  # Skip the length field

        unknown_items: list[tuple[int, bytes]] = []
        values = _get_default_field_values(cls)

        while stream.tell() < len(data):
            item_id = int.from_bytes(stream.read(ITEM_ID_BYTES))
            item_length = int.from_bytes(stream.read(ITEM_LENGTH_BYTES))
            item_data = stream.read(item_length)

            for command_field in fields(cls):
                if get_origin(command_field.type) is not Annotated:
                    continue

                metadata = command_field.type.__metadata__
                if len(metadata) != 1 or not isinstance(metadata[0], Item):
                    continue

                metadata = metadata[0]

                if metadata.item_id != item_id:
                    continue

                current_value = values[command_field.name]
                # If the field is a collection, append the deserialized item to the current value via restructuring
                if isinstance(current_value, Collection) and not isinstance(current_value, (str, bytes)):
                    if command_field.default_factory is MISSING:
                        msg = "Collection fields must have a default factory!"
                        raise ValueError(msg)

                    values[command_field.name] = command_field.default_factory(
                        (
                            *current_value,
                            *command_field.default_factory((metadata.transformer.deserialize(item_data),)),
                        ),
                    )

                # If the field is not a collection, set the value directly — this will overwrite any previous value
                else:
                    values[command_field.name] = metadata.transformer.deserialize(item_data)

        # noinspection PyArgumentList
        return cls(**values, unknown_items=unknown_items)

    def __repr__(self: Self) -> str:
        """Similar to the default repr, but with tabs and newlines for readability and excludes None values."""
        result = f"{self.__class__.__name__}(\n\t"

        attrs = [f"{attr.name}={value!r}" for attr in fields(self) if (value := getattr(self, attr.name)) is not None]

        result += ",\n\t".join(attrs).strip() + "\n)"
        return result

    def bytes_debug_repr(self: Self) -> str:
        """Return a debug-focused representation of the command's raw bytes."""
        stream = BytesIO(bytes(self))
        command_id = stream.read(COMMAND_ID_BYTES)
        command_length = stream.read(COMMAND_LENGTH_BYTES)
        result = (
            f"COMMAND_ID = {command_id} (0x{self.command_id:02X})\n"
            f"COMMAND_LENGTH = {command_length} ({int.from_bytes(command_length)})"
        )

        while stream.tell() < len(bytes(self)):
            item_id_bytes = stream.read(ITEM_ID_BYTES)
            item_id = int.from_bytes(item_id_bytes)

            item_length_bytes = stream.read(ITEM_LENGTH_BYTES)
            item_length = int.from_bytes(item_length_bytes)

            item_data = stream.read(item_length)

            result += (
                "\n"
                f"ITEM_ID = {item_id_bytes} (0x{item_id:02X})\n"
                f"\tITEM_LENGTH = {item_length_bytes} ({item_length})\n"
                f"\tITEM_DATA = {item_data}\n"
            )

        return result



#  Copyright (C) 2024  Cypheriel
from pathlib import Path

from appdirs import AppDirs

_APP_DIRS = AppDirs(
    "PommeKit",
    "Cypheriel",
)

USER_DATA_DIR = Path(_APP_DIRS.user_data_dir)
USER_LOG_DIR = Path(_APP_DIRS.user_log_dir)



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel

from __future__ import annotations

import hmac
import logging
import plistlib
from base64 import b64encode
from dataclasses import dataclass
from enum import StrEnum
from hashlib import pbkdf2_hmac, sha256
from typing import TYPE_CHECKING, Final, Self

from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import CBC
from cryptography.hazmat.primitives.padding import PKCS7
from httpx import AsyncClient

from ..gsa._srp import SRPUser
from ._exceptions import GrandSlamError

if TYPE_CHECKING:
    from collections.abc import Callable

GSA_BASE_URL: Final = "https://gsa.apple.com"
GSA_SERVICE_URL: Final = f"{GSA_BASE_URL}/grandslam/GsService2"
GSA_TRUSTED_DEVICE_URL: Final = f"{GSA_BASE_URL}/auth/verify/trusteddevice"
GSA_VALIDATE_2FA_URL: Final = f"{GSA_SERVICE_URL}/validate"

APPLE_PLIST_HEADER: Final = (
    b"<?xml version='1.0' encoding='UTF-8'?>\n"
    b"<!DOCTYPE plist PUBLIC '-//Apple//DTD PLIST 1.0//EN' 'https://www.apple.com/DTDs/PropertyList-1.0.dtd'>"
)

APPLE_REJECT_DEVICE_ERROR: Final = -22421

logger = logging.getLogger(__name__)


@dataclass
class GSAResponseStatus:
    status_code: int
    status_message: str | None

    @property
    def is_success(self: Self) -> bool:
        return self.status_code == 0


def _response_status(response: dict[str, ...]) -> GSAResponseStatus:
    status = response.get("Status", response)
    status_code = status["ec"]

    return GSAResponseStatus(status_code, status.get("em"))


def _verify_headers(headers: dict[str, str]) -> bool:
    logger.debug(f"Verifying headers: {headers}")
    expected_headers = {
        h.lower()
        for h in (
            "User-Agent",
            "X-Apple-I-MD",
            "X-Apple-I-MD-RINFO",
            "X-Apple-I-MD-LU",
            "X-Apple-I-MD-M",
            # "X-Apple-I-SRL-NO",
            "X-Apple-I-Client-Time",
            "X-Apple-I-TimeZone",
            "X-Apple-Locale",
            "X-Mme-Client-Info",
            "X-Mme-Device-Id",
        )
    }

    missing_headers = expected_headers - {h.lower() for h in headers}
    if missing_headers:
        logger.warning(f"Missing headers: {missing_headers}")

    return len(missing_headers) == 0


class AUStatus(StrEnum):
    UNKNOWN = "UNKNOWN"
    TRUSTED_DEVICE = "trustedDeviceSecondaryAuth"
    SMS = "secondaryAuth"
    COMPLETE = "COMPLETE"
    NOT_STARTED = "NOT_STARTED"

    @classmethod
    def _missing_(cls: type[Self], _value: str) -> AUStatus:
        return cls.UNKNOWN


class GSAClient:
    def __init__(self: Self, anisette_provider: Callable[[], dict[str, str]]) -> None:
        """Initialize a new instance of the GSAClient class."""
        self.pet_token: str | bytes | None = None
        self.au_status: AUStatus = AUStatus.NOT_STARTED

        self._client = AsyncClient(
            verify=False,  # noqa: S501
        )
        self._anisette_provider = anisette_provider

        self._apple_id: str | None = None
        self._srp_user: SRPUser | None = None

        self._identity_token: str | None = None
        self._extra_headers: dict[str, str] = {}

    async def _request(self: Self, data: dict[str, ...]) -> dict[str, ...]:
        """Send a privileged request to the GrandSlam service."""
        headers = self._anisette_provider()
        if _verify_headers(headers) is False:
            msg = "Invalid Anisette headers."
            raise ValueError(msg)

        payload = {
            "Header": {"Version": "1.0.1"},
            "Request": {"cpd": headers} | data,
        }

        response = await self._client.post(GSA_SERVICE_URL, content=plistlib.dumps(payload), headers=headers)
        if not response.is_success:
            msg = f"Failed to authenticate with GrandSlam: {response}"
            raise GrandSlamError(msg)

        response_data = plistlib.loads(response.content)["Response"]

        status = _response_status(response_data)

        if status.status_code == APPLE_REJECT_DEVICE_ERROR:
            msg = "Apple rejected machine data. Please re-provision this device."
            raise GrandSlamError(msg, error_code=status.status_code, error_message=status.status_message)

        if status.is_success is False:
            logger.debug(f"{response_data = }")
            msg = "Unexpected error occurred!"
            raise GrandSlamError(msg, error_code=status.status_code, error_message=status.status_message)

        return response_data

    def _get_session_key(self: Self, name: str) -> bytes:
        """Generate a session key using the SRP session key."""
        return hmac.new(self._srp_user.session_key, name.encode(), sha256).digest()

    def _decrypt_cbc(self: Self, cbc_data: bytes) -> bytes:
        """Decrypt the CBC data using session keys."""
        extra_data_key = self._get_session_key("extra data key:")
        extra_data_iv = self._get_session_key("extra data iv:")[:16]

        cipher = Cipher(AES(extra_data_key), CBC(extra_data_iv))

        decryptor = cipher.decryptor()
        data = decryptor.update(cbc_data) + decryptor.finalize()

        unpadder = PKCS7(128).unpadder()
        return unpadder.update(data) + unpadder.finalize()

    async def request_device_2fa(self: Self) -> None:
        """Request a trusted device 2FA code from Apple."""
        self._extra_headers = {
            "User-Agent": "Xcode",
            "Accept": "text/x-xml-plist",
            "X-Apple-Identity-Token": self._identity_token,
        }

        response = await self._client.get(
            GSA_TRUSTED_DEVICE_URL,
            headers=self._anisette_provider() | self._extra_headers,
        )

        logger.debug(f"{response.request.headers = }")

        if not response.is_success:
            logger.debug(f"{response.content = }")
            msg = "Failed to request trusted device 2FA."
            raise GrandSlamError(
                msg,
                error_code=response.status_code,
                error_message=response.text,
            )

    async def validate_device_2fa(self: Self, security_code: str) -> bool:
        """Validate the trusted device 2FA code with Apple."""
        self._extra_headers["security-code"] = security_code
        anisette_headers = self._anisette_provider()
        _verify_headers(anisette_headers | self._extra_headers)

        response = await self._client.get(GSA_VALIDATE_2FA_URL, headers=anisette_headers | self._extra_headers)
        logger.debug(f"{response.http_version} {response.status_code} {response.reason_phrase}")
        logger.debug("\n".join(f"{k}: {v}" for k, v in response.headers.items()))
        logger.debug(f"{response.content = }")

        response_data = plistlib.loads(response.content)

        status = _response_status(response_data)

        if status.is_success is False:
            logger.error(f"Failed to validate trusted device ({status.status_code}: {status.status_message})")

        return status.is_success

    async def authenticate(self: Self, apple_id: str, password: str) -> None:
        """Authenticate with GrandSlam using the provided Apple ID and password."""
        self._apple_id = apple_id
        self._srp_user = SRPUser(self._apple_id)

        payload = {
            "A2k": self._srp_user.public_ephemeral,
            "ps": ["s2k", "s2k_fo"],
            "u": self._apple_id,
            "o": "init",
        }

        logger.info("Starting authentication with GrandSlam.")
        logger.debug(f"{payload = }")

        init_response = await self._request(payload)

        logger.debug(f"{init_response = }")

        password_hmac = pbkdf2_hmac(
            "sha256",
            sha256(password.encode()).digest(),
            init_response["s"],
            init_response["i"],
        )[:32]

        complete_response = await self._request(
            {
                "c": init_response["c"],
                "M1": self._srp_user.process_challenge(password_hmac, init_response["s"], init_response["B"]),
                "u": self._apple_id,
                "o": "complete",
            },
        )

        logger.debug("Sending confirmation response with proof to GrandSlam.")
        logger.debug(f"{complete_response = }")

        if not self._srp_user.verify_session(complete_response["M2"]):
            logger.debug(f"{complete_response = }")
            msg = "Failed to verify user SRP session."
            raise GrandSlamError(msg)

        spd_data = self._decrypt_cbc(complete_response["spd"])
        spd = plistlib.loads(APPLE_PLIST_HEADER + spd_data)

        au = complete_response.get("Status", {}).get("au")
        self.au_status = AUStatus.COMPLETE if au is None else AUStatus(au)

        match self.au_status:
            case AUStatus.TRUSTED_DEVICE:
                logger.warning("Trusted device secondary authentication required.")
            case AUStatus.SMS:
                logger.warning("Secondary authentication required.")
            case AUStatus.UNKNOWN:
                logger.warning("Unknown authentication status.")
            case AUStatus.COMPLETE:
                logger.info("Authentication complete.")

        self.pet_token = spd.get("t", {}).get("com.apple.gs.idms.pet", {}).get("token", None)
        self._identity_token = b64encode(f"{spd['adsid']}:{spd['GsIdmsToken']}".encode()).decode()



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel

"""Module containing the Albert activation logic."""

from __future__ import annotations

import plistlib
import re
from importlib import resources
from logging import getLogger
from typing import TYPE_CHECKING, Final, Literal, TypedDict

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.hashes import SHA1
from cryptography.hazmat.primitives.serialization import Encoding, load_pem_private_key
from cryptography.x509 import Certificate, load_pem_x509_certificate
from httpx import AsyncClient, Response

from ..albert.device_csr import generate_device_csr

if TYPE_CHECKING:
    from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

_RESOURCE_ROOT: Final = resources.files(__package__)
FAIRPLAY_PRIVATE_KEY = load_pem_private_key((_RESOURCE_ROOT / "fairplay.key").read_bytes(), password=None)
FAIRPLAY_CERT_CHAIN = (_RESOURCE_ROOT / "fairplay-chain.crt").read_bytes()

ACTIVATION_URL: Final = "https://albert.apple.com/deviceservices/deviceActivation?device=MacOS"
CERTIFICATE_RESPONSE_PATTERN: Final = re.compile(r"<Protocol>(.*)</Protocol>")


class ActivationInfoContent(TypedDict):
    """The content of the activation info."""

    ActivationRandomness: str
    ActivationState: Literal["Unactivated"]
    DeviceCertRequest: bytes
    DeviceClass: Literal["MacOS", "Windows"]
    ProductType: str
    ProductVersion: str
    BuildVersion: str
    SerialNumber: str
    UniqueDeviceID: str


class ActivationInfo(TypedDict):
    """The activation info sent to Apple through Albert."""

    ActivationInfoComplete: Literal[True]
    ActivationInfoXML: bytes
    FairPlayCertChain: bytes
    FairPlaySignature: bytes


ActivationPayload = TypedDict("ActivationPayload", {"activation-info": str})

logger = getLogger(__name__)


async def request_push_certificate(
    private_key: RSAPrivateKey | None = None,
    activation_info_content: ActivationInfoContent | None = None,
    fairplay_private_key: RSAPrivateKey = FAIRPLAY_PRIVATE_KEY,
    fairplay_cert_chain: bytes = FAIRPLAY_CERT_CHAIN,
    http_client: AsyncClient | None = None,
) -> tuple[RSAPrivateKey, Certificate]:
    """
    Request a push certificate from Albert.

    :param private_key:
    :param activation_info_content: The activation info content to send to Albert.
    :param fairplay_cert_chain: Optional FairPlay certificate chain override.
    :param fairplay_private_key: Optional FairPlay private key override.
    :param http_client: The optional async HTTP _client to use for the request.
    :return: A `tuple` containing the private key and the push certificate.
    """
    if private_key is None:
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    if "DeviceCertRequest" not in activation_info_content:
        activation_info_content["DeviceCertRequest"] = generate_device_csr(private_key).public_bytes(Encoding.PEM)

    activation_info_xml = plistlib.dumps(activation_info_content)
    activation_signature = fairplay_private_key.sign(activation_info_xml, PKCS1v15(), SHA1())  # noqa: S303

    activation_info: ActivationInfo = {
        "ActivationInfoComplete": True,
        "ActivationInfoXML": activation_info_xml,
        "FairPlayCertChain": fairplay_cert_chain,
        "FairPlaySignature": activation_signature,
    }

    payload: ActivationPayload = {"activation-info": plistlib.dumps(activation_info).decode()}

    match http_client:
        case AsyncClient() | None:
            async with http_client or AsyncClient() as client:
                response: Response = await client.post(
                    ACTIVATION_URL,
                    data=payload,
                )
        case _:
            msg = "Unsupported HTTP _client provided."
            raise ValueError(msg)

    if not response.is_success:
        msg = f"Received non-OK status code: {response.status_code}"
        raise RuntimeError(msg)

    if (match := CERTIFICATE_RESPONSE_PATTERN.search(response.text)) is None:
        logger.debug(f"Failed to extract certificate from response: {response.text}")
        msg = "Failed to extract certificate from response."
        raise RuntimeError(msg)

    protocol_data = plistlib.loads(match.group(1).encode())
    certificate_data = protocol_data["device-activation"]["activation-record"]["DeviceCertificate"]

    device_certificate = load_pem_x509_certificate(certificate_data)
    return private_key, device_certificate



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""APNsServerStream class definition."""

from __future__ import annotations

import asyncio
import logging
from asyncio import Server, StreamReader, StreamWriter
from logging import getLogger
from typing import TYPE_CHECKING, Awaitable, Callable, Self

from ..streams._apns_stream import ConnectionType, _APNsStream

if TYPE_CHECKING:
    from types import TracebackType

    from ..protocol.packet import APNsCommand


class APNsServerStream(_APNsStream):
    """Stream class for server APNs courier hosts."""

    def __init__(self: Self, host: str = "localhost", port: int = 5223) -> None:
        """Initialize a new APNs server stream."""
        super().__init__(ConnectionType.SERVER, host, port)

        self._server: Server | None = None
        self._handle_client: Callable[[APNsServerStream, StreamReader, StreamWriter], Awaitable[None]] | None = None
        self._logger = getLogger(f"{__name__}.{self.__class__.__name__}")

    async def __aenter__(self: Self) -> Self:
        """Async context manager enter method."""
        await self.start()
        return self

    async def __aexit__(
        self: Self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Async context manager exit method."""
        await self.close()

    def log(
        self: Self,
        level: int,
        msg: str,
        *args: str,
        client_address: tuple[str, int] | None = None,
        **kwargs: ...,
    ) -> None:
        """Log a message with the server's logger."""
        addr_repr = f"[{client_address[0]}:{client_address[1]}] " if client_address else ""
        self._logger.log(level, f"Server :: {addr_repr}{msg}", *args, **kwargs, stacklevel=2)

    async def start(self: Self, server: Server | None = None) -> None:
        """Start the server."""
        self._server = server or await asyncio.start_server(
            self._wrapped_callback,
            self.host,
            self.port,
            ssl=self._create_ssl_context(ConnectionType.SERVER),
        )

        self.log(logging.DEBUG, f"Listening on {self.host}:{self.port} for APNs connections!")

    async def close(self: Self) -> None:
        """Close the server."""
        self._server.close()
        await self._server.wait_closed()

    async def wait_closed(self: Self) -> None:
        """Wait for the server to close."""
        await self._server.wait_closed()

    def set_callback(
        self: Self,
        callback: Callable[[APNsServerStream, StreamReader, StreamWriter], Awaitable[None]],
    ) -> None:
        """Set the callback to handle client connections."""
        self._handle_client = callback

    async def _wrapped_callback(self: Self, stream_reader: StreamReader, stream_writer: StreamWriter) -> None:
        """Wrap the callback to handle logging and other tasks."""
        client_host, client_port, *_ = stream_writer.get_extra_info("peername")
        self.log(logging.DEBUG, "Received client connection.", client_address=(client_host, client_port))

        await self._handle_client(self, stream_reader, stream_writer)

        self.log(logging.DEBUG, "Finished handling client connection.", client_address=(client_host, client_port))

    async def send(self: Self, stream_writer: StreamWriter, command: APNsCommand) -> None:
        """Send a command to the stream."""
        addr_info = stream_writer.get_extra_info("peername")

        self.log(
            logging.DEBUG,
            f"Sending {command.__class__.__name__} (0x{command.command_id:02X}).",
            client_address=addr_info,
        )
        await self._send_command(stream_writer, command)

    async def read(self: Self, stream_reader: StreamReader) -> APNsCommand:
        """Read a command from the stream."""
        return await self._read_command(stream_reader)



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
from __future__ import annotations

from typing import Self


class GrandSlamError(Exception):
    def __init__(
        self: Self,
        message: str = "",
        error_code: int | None = None,
        error_message: str = "Error message not supplied.",
    ) -> None:
        """Exception raised when an error occurs during a GrandSlam operation."""
        status_message = ""

        if error_code is not None:
            status_message = f" ({error_code}: {error_message})"

        super().__init__(f"{message}{status_message}")



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""Package containing a client implementation for Apple's GrandSlam Authentication (GSA) service."""

from ._client import AUStatus, GSAClient

__all__ = [
    "AUStatus",
    "GSAClient",
]



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""
Simple mapping of known topics to their hashed values.

>>> TOPIC_HASHES[sha1(b"com.apple.madrid", usedforsecurity=False).digest()] == "com.apple.madrid"
True
"""

from __future__ import annotations

from hashlib import sha1
from typing import Final

_KNOWN_TOPICS: Final = frozenset(
    {
        "com.apple.madrid",
    },
)
TOPIC_HASHES: Final[dict[bytes, str]] = {
    sha1(topic.encode(), usedforsecurity=False).digest(): topic for topic in _KNOWN_TOPICS
}



#  Copyright (C) 2024  Cypheriel
import asyncio
from collections.abc import Callable, Coroutine
from functools import wraps


def run_async(func: Callable[..., Coroutine]) -> Callable:
    @wraps(func)
    def wrapper(*args: ..., **kwargs: ...) -> ...:
        return asyncio.run(func(*args, **kwargs))

    return wrapper



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""Status codes used by Apple's various services."""

from __future__ import annotations

from enum import IntEnum


class AppleStatusCode(IntEnum):
    """List of status codes used by Apple's various services."""

    SUCCESS = 0
    NO_RESULT_CODE = 1
    LEGACY_TUNNEL_OK = 200
    LEGACY_TUNNEL_BAD_REQUEST = 400
    PARSE_FAILURE = 1000
    INVALID_FIELD = 1001
    MESSAGE_REMOVED = 1002
    MESSAGE_INHIBITED = 1003
    DECODE_FAILURE_BASE_64 = 1004
    DECODE_FAILURE_HEX = 1005
    AUTH_FAILURE = 1006
    STORAGE_FAILURE = 1007
    INTERNAL_ERROR = 5001
    SERVER_TOO_BUSY = 5004
    BAD_PUSH_TOKEN = 5006
    MISSING_REQUIRED_KEY = 5008
    UPLUSONE_DISABLED = 5010
    BAD_SIGNATURE = 5032
    WEB_TUNNEL_SERVICE_DISABLED = 5200
    WEB_TUNNEL_SERVICE_INTERNAL_ERROR = 5201
    WEB_TUNNEL_SERVICE_TOO_BUSY = 5202
    WEB_TUNNEL_SERVICE_MISSING_KEY = 5203
    WEB_TUNNEL_SERVICE_MALFORMED_URL = 5204
    WEB_TUNNEL_SERVICE_UNAUTHORIZED_URL = 5205
    WEB_TUNNEL_SERVICE_RESPONSE_TOO_LARGE = 5206
    WEB_TUNNEL_SERVICE_ORIGIN_SERVER_TIMEOUT = 5207
    UNKNOWN_COMMAND = 6000
    RATE_LIMITED = 7000
    SERVICE_DISABLED = 7001
    ALIAS_EMPTY = 5035
    EMAIL_INVALID = 5036
    EMAIL_IN_REQUIRED = 5037
    EMAIL_IN_USE = 5038
    EMAIL_IN_USE_ALT = 5055
    EMAIL_LEGACY_OR_INACTIVE = 5088
    EMAIL_VETTED = 5051
    EMAIL_NOT_VETTED = 5052
    HTTP_SUCCESS = 200
    BAD_REQUEST = 5003
    CANCELLED_BY_USER = 5013
    UNSUPPORTED_MANAGED_ID = 5100
    UNSUPPORTED_AUDITOR_ID = 5103
    VALIDATION_FAILURE = 8000
    VALIDATION_CERT_FETCH_FAILURE = 8001
    NO_NETWORK_AVAILABLE = 10000
    FAILED_REQUEST = 20000
    PUSH_REQUEST_FAILED = 20001
    ACTION_DO_NOT_RETRY = 6001
    ACTION_RETRY_WITH_DELAY = 6002
    ACTION_RETRY_WITH_CORRECTED_TIME = 6003
    ACTION_RETRY_WITH_NEW_ABSINTHE_CONTEXT = 6004
    ACTION_REFRESH_CREDENTIALS = 6005
    ACTION_REFRESH_REGISTRATION = 6006
    ACTION_AUTHENTICATION_FAILED = 6008
    ACTION_PERMANENT_FAILURE = 6009
    ACTION_ID_MS_SESSION_REFRESH_EXCEPTION = 6010
    ACTION_PROVISION_LIMIT_REACHED = 6011
    ACTION_PROVISION_REPLACE_ALL_SESSION_KEYS = 6012
    UNAUTHENTICATED = 5000
    INVALID_NAME_OR_PASSWORD = 5012
    ACCOUNT_DISABLED_FOR_SECURITY_REASONS = 5046
    VERSION_MISMATCH = 5002
    NO_SUCH_PLAYER_ID = 5005
    NO_SUCH_ACHIEVEMENT = 5007
    MISSING_REQUIRED_HEADER = 5009
    UNRECOGNIZED_GAME_DESCRIPTOR_HEADERS = 5010
    MUST_ACCEPT_TERMS_OF_SERVICE = 5011
    NO_SUCH_FRIEND_RELATIONSHIP = 5014
    NO_SUCH_ALIAS = 5015
    EMAIL_NOT_IN_NETWORK = 5016
    REQUEST_DISALLOWED = 5017
    NO_SUCH_RID = 5018
    NO_SUCH_GAME = 5019
    MIXED_RESPONSE = 5020
    PROFILE_EXISTS = 5021
    ALIAS_EXISTS = 5022
    ACCOUNT_NAME_NOT_EMAIL_ADDRESS = 5023
    ALIAS_TOO_LONG = 5024
    ALIAS_TOO_FREQUENT_UPDATES = 5025
    PLAYER_STATUS_NOT_ALLOWED = 5026
    PLAYER_STATUS_TOO_LONG = 5027
    NO_SUCH_BUCKET_ID = 5028
    AUTHENTICATED_NOT_AUTHORIZED = 5029
    NO_PUSH_TOKEN_FOR_ID = 5030
    PUSH_PAYLOAD_TOO_BIG = 5031
    NEW_SIGNATURE_REQUIRED = 5033
    NEW_LINKED_SIGNATURE_REQUIRED = 5099
    NEW_REGISTRATION_REQUIRED = 5034
    EXPIRED_RELAY_TOKEN = 5039
    BAD_SESSION_TOKEN = 5040
    EXPIRED_SESSION_TOKEN = 5041
    FORCE_PASSWORD_CHANGE = 5047
    PROFILE_BLOCKLISTED = 5048
    SELF_VERSION_TOO_OLD = 5057
    PEER_VERSION_TOO_OLD = 5058
    PHONE_NUMBER_TOO_SHORT = 5059
    PHONE_NUMBER_TOO_LONG = 5060
    PHONE_NUMBER_BAD_COUNTRY_CODE = 5061
    PHONE_NUMBER_MALFORMED = 5062
    BAD_INVITATION_CONTEXT = 5063
    UNPROMOTABLE_PHONE_NUMBER = 5064
    REGISTRATION_LIMIT_REACHED = 5068
    ABSINTHE_INTERNAL_ERROR = 5078
    ABSINTHE_SESSION_CREATION_FAILED = 5079
    ABSINTHE_VALIDATION_FAILED = 5080
    ALBERT_INTERNAL_ERROR = 5089
    ALBERT_VALIDATION_FAILED = 5090
    NEEDS_BREAK_BEFORE_MAKE = 5076
    BAD_CERT = 5085
    BAD_NONCE = 5086
    BAD_NONCE_TIMESTAMP = 5087
    SERVER_REGISTRATION_UNSUPPORTED = 5092
    NONE = 255

    UNKNOWN = -1

    def __str__(self: AppleStatusCode) -> str:
        """Return a human-readable representation of the status code."""
        return f"{self.name} ({self.value})"

    @classmethod
    def _missing_(cls: type[AppleStatusCode], value: object) -> AppleStatusCode:
        """Return the `UNKNOWN` status code for any unknown value."""
        del value
        return cls.UNKNOWN



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
from __future__ import annotations

import ssl
from enum import Enum, auto
from importlib import resources
from logging import getLogger
from ssl import SSLContext
from tempfile import TemporaryFile
from typing import TYPE_CHECKING, Final, Self

from ..protocol.commands import COMMAND_MAP, Direction

if TYPE_CHECKING:
    from asyncio import StreamReader, StreamWriter
    from importlib.abc import Traversable

    from ..protocol.packet import APNsCommand

_RESOURCE_ROOT = resources.files(__package__)
_SERVER_CERTIFICATE = _RESOURCE_ROOT / "server.crt"
_SERVER_PRIVATE_KEY = _RESOURCE_ROOT / "server.key"

ALPN_PROTOCOL: Final = ("apns-security-v3",)

logger = getLogger(__name__)


class ConnectionType(Enum):
    CLIENT = auto()
    SERVER = auto()


class _APNsStream:
    def __init__(self: Self, connection_type: ConnectionType, host: str | None = None, port: int = 5223) -> None:
        """Initialize a new APNs stream."""
        self.host = host
        self.port = port
        self.connection_type = connection_type

        if self.connection_type not in {ConnectionType.CLIENT, ConnectionType.SERVER}:
            msg = f"Invalid direction for APNs stream: {self.connection_type}"
            raise ValueError(msg)

    @staticmethod
    def _create_ssl_context(
        connection_type: ConnectionType,
        cert_file_path: Traversable | None = None,
        key_file_path: Traversable | None = None,
    ) -> SSLContext:
        """Create an SSL context for the APNs connection."""
        if cert_file_path is None and key_file_path is None:
            cert_file = TemporaryFile()
            cert_file.write(_SERVER_CERTIFICATE.read_bytes())

            key_file = TemporaryFile()
            key_file.write(_SERVER_PRIVATE_KEY.read_bytes())
        elif cert_file_path is not None and key_file_path is not None:
            cert_file = cert_file_path
            key_file = key_file_path
        else:
            msg = "Both certificate and private key paths must be provided."
            raise ValueError(msg)

        match connection_type:
            case ConnectionType.CLIENT:
                ssl_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            case ConnectionType.SERVER:
                ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
                ssl_context.load_cert_chain(certfile=cert_file.name, keyfile=key_file.name)
                ssl_context.options |= ssl.OP_SINGLE_DH_USE
                ssl_context.options |= ssl.OP_SINGLE_ECDH_USE
            case _:
                msg = f"Invalid connection type: {connection_type}"
                raise ValueError(msg)

        ssl_context.set_alpn_protocols(ALPN_PROTOCOL)
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE

        return ssl_context

    @staticmethod
    async def _read_packet(stream_reader: StreamReader) -> bytes:
        """Read a packet from the stream."""
        command_id = await stream_reader.readexactly(1)
        command_length = await stream_reader.readexactly(4)
        items_data = await stream_reader.readexactly(int.from_bytes(command_length))
        return command_id + command_length + items_data

    async def _read_command(self: Self, stream_reader: StreamReader) -> APNsCommand:
        """Read an `APNsCommand`-deserialized object from the stream."""
        packet = await self._read_packet(stream_reader)

        match self.connection_type:
            case ConnectionType.CLIENT:
                direction = Direction.CLIENT_BOUND
            case ConnectionType.SERVER:
                direction = Direction.SERVER_BOUND
            case _:
                msg = f"Invalid connection type: {self.connection_type}"
                raise ValueError(msg)

        for (command_id, command_direction), command_type in COMMAND_MAP.items():
            if command_id == packet[0] and direction in command_direction:
                return command_type.from_bytes(packet, includes_header=True)

        msg = f"Failed to find command for packet: {packet!r} ({direction})"
        raise ValueError(msg)

    @staticmethod
    async def _send_packet(stream_writer: StreamWriter, packet: bytes) -> None:
        """Send a packet to the stream."""
        stream_writer.write(packet)
        await stream_writer.drain()

    async def _send_command(self: Self, stream_writer: StreamWriter, command: APNsCommand) -> None:
        """Send an `APNsCommand`, bytes-serialized, object to the stream."""
        await self._send_packet(stream_writer, bytes(command))



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""Module for the APNs client and server streams."""

from .client_stream import APNsClientStream
from .server_stream import APNsServerStream

__all__ = ["APNsClientStream", "APNsServerStream"]



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel

"""Module containing the high-level APNs client based on the APNs event listener."""

from __future__ import annotations

import asyncio
from asyncio import Event, IncompleteReadError
from logging import getLogger
from typing import TYPE_CHECKING, Self

from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.hashes import SHA1

from .._util.event_listener import EventType
from ..apns.listener import APNsListener
from ..apns.protocol.commands import (
    ClientBoundPushNotificationCommand,
    ConnectCommand,
    ConnectResponseCommand,
    FilterTopicsCommand,
    KeepAliveCommand,
)
from ..apns.protocol.transformers import TOPIC_TRANSFORMER, Interface, Nonce, Status, UnknownFlag
from ..apns.streams import APNsClientStream

if TYPE_CHECKING:
    from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
    from cryptography.x509 import Certificate

logger = getLogger(__name__)


class APNsClient(APNsListener):
    """High-level APNs client based on the APNs event listener."""

    def __init__(
        self: Self,
        push_token: str | None = None,
        push_key: RSAPrivateKey | None = None,
        push_cert: Certificate | None = None,
        flags: UnknownFlag = UnknownFlag.IS_ROOT,
        interface: Interface = Interface.WIFI,
        carrier: str = "WiFi",
        os_version: str = "10.6.4",
        os_build: str = "10.6.4",
        hardware_version: str = "windows1,1",
        nonce: Nonce = None,
        courier_address: str | None = None,
        courier_port: int = 5223,
    ) -> None:
        """
        Initialize the APNs client.

        :param push_token: The optional push token returned by a prior APNs connection
        :param push_key: The private key used to sign the CSR sent to Albert
        :param push_cert: The push certificate returned by Albert
        :param flags: Device flags — most flags are unknown and undocumented for the time being
        :param interface: The connection interface — Wi-Fi or cellular
        :param carrier: The carrier name or "Wi-Fi" if using Wi-Fi
        :param os_version: The OS version
        :param os_build: The OS build version specifier
        :param hardware_version: The hardware version E.g. "iPhone1,1"
        :param nonce: The nonce used in creation of the signature
        :param courier_address: The address of the APNs courier
        :param courier_port: The courier port
        """
        super().__init__()

        if courier_address is None:
            courier_address = APNsClientStream.fetch_random_courier_address()

        self.courier_address = courier_address
        self.courier_port = courier_port

        self.push_key = push_key
        self.push_cert = push_cert
        self.push_token = push_token

        self.flags = flags
        self.interface = interface
        self.carrier = carrier
        self.os_version = os_version
        self.os_build = os_build
        self.hardware_version = hardware_version
        self.nonce = nonce or Nonce()

        self.signature: bytes | None = None
        self.on_connected_event: Event = Event()

        self._apns_stream: APNsClientStream | None = None

        self._register_default_listeners()

    def _register_default_listeners(self: Self) -> None:
        @self._register_command_listener(ConnectResponseCommand)
        async def _on_connect_response(command: ConnectResponseCommand) -> None:
            if command.status is Status.ERROR:
                logger.error("Received ConnectResponseCommand (0x08) with status ERROR (0x02)!")
                await self.close()
                return

            if command.push_token is not None:
                self.push_token = command.push_token

            self.on_connected_event.set()

            # NOTE: This event should be triggered *after* internal post-connection processing.
            await self._trigger_event(EventType.CONNECT, EventType.CONNECT.name)

        @self._register_command_listener(ClientBoundPushNotificationCommand)
        async def _on_push_notification(command: ClientBoundPushNotificationCommand) -> None:
            topic = TOPIC_TRANSFORMER.serialize(command.topic)

            await self._trigger_event(EventType.TOPIC, topic, self, command)

    async def _connect(self: Self) -> None:
        self._apns_stream = APNsClientStream(self.courier_address, self.courier_port)
        await self._apns_stream.connect()

        connect_command = ConnectCommand(
            push_token=self.push_token,
            state=b"\x01",
            flags=self.flags,
            interface=self.interface,
            carrier=self.carrier,
            os_version=self.os_version,
            os_build=self.os_build,
            hardware_version=self.hardware_version,
            certificate=self.push_cert,
            nonce=self.nonce,
            signature=self.signature,
            protocol_version__=2,
            redirect_count=0,
        )

        logger.debug(f"Sending connect command: {connect_command}")
        logger.debug(f"{connect_command.bytes_debug_repr()}")

        await self._apns_stream.send(connect_command)

    async def filter_topics(self: Self) -> None:
        """Send a FilterTopicsCommand to the APNs courier to filter push notifications to ``self.enabled_topics``."""
        await self._apns_stream.send(
            FilterTopicsCommand(push_token=self.push_token, enabled_topics=self.enabled_topics),
        )

    async def run(self: Self) -> None:
        """Start the APNs client, connect to the courier, filter topics, and dispatch incoming packets to events."""
        if self.push_key is None:
            logger.warning("No push key was provided.")

        if self.push_cert is None:
            logger.warning("No push certificate was provided.")

        if self.push_key is not None and self.push_cert is not None:
            self.signature = b"\x01\x01" + self.push_key.sign(bytes(self.nonce), PKCS1v15(), SHA1())  # noqa: S303

        await self._connect()

        task = asyncio.create_task(self.on_connected_event.wait())
        task.add_done_callback(lambda _: asyncio.create_task(self.filter_topics()))

        while self._apns_stream is not None:
            try:
                command = await self._apns_stream.read()

            except IncompleteReadError:
                await self._apns_stream.send(KeepAliveCommand())
                return

            _results = await self._trigger_event(EventType.COMMAND, command.__class__, command)

    async def close(self: Self) -> None:
        """Close the APNs stream's connection to the APNs courier."""
        if self._apns_stream is not None:
            await self._apns_stream.close()
            self._apns_stream = None



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""Module containing serializers and deserializers for APNs protocol fields."""

from __future__ import annotations

from base64 import b64decode, b64encode
from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import IntEnum, IntFlag, auto
from functools import lru_cache, partial
from hashlib import sha1
from io import BytesIO
from typing import Callable, Final, Generic, Literal, Self, SupportsIndex, TypeVar

from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.x509 import load_der_x509_certificate

from ..._util.crypto import randbytes
from ..protocol.topics import TOPIC_HASHES

__all__ = [
    "CAPABILITIES_TRANSFORMER",
    "DER_CERTIFICATE_TRANSFORMER",
    "INTERFACE_TRANSFORMER",
    "MS_DATETIME_TRANSFORMER",
    "NONCE_TRANSFORMER",
    "NO_OP_TRANSFORMER",
    "PUSH_TOKEN_TRANSFORMER",
    "STATUS_TRANSFORMER",
    "STRING_TRANSFORMER",
    "TOPIC_TRANSFORMER",
    "UNKNOWN_FLAG_TRANSFORMER",
    "Capability",
    "DataTransformer",
    "Interface",
    "Nonce",
    "Status",
    "UnknownFlag",
    "integer_transformer",
]

_T = TypeVar("_T")


def reverse_bits(data: bytes, byte_length: int = 4) -> bytes:
    """Reverse the bits of a byte sequence."""
    return int(f"{int.from_bytes(data):>0{byte_length * 8}b}"[::-1], 2).to_bytes(length=byte_length)


class DataTransformer(Generic[_T]):
    """A class responsible for serializing and deserializing bytes data."""

    def __init__(self: Self, deserializer: Callable[[bytes], _T], serializer: Callable[[_T], bytes]) -> None:
        """
        Initialize a new data transformer.

        :param deserializer: The function to deserialize bytes data.
        :param serializer: The function to serialize data to bytes.
        """
        self.deserialize = deserializer
        self.serialize = serializer


class BitMask(IntFlag):
    """Simple wrapper class to serialize and deserialize ``IntFlag`` in reverse bit order."""

    @classmethod
    def from_bytes(
        cls: type[Self],
        bytes_: bytes,
        byteorder: Literal["big", "little"] = "big",
        *,
        signed: bool = False,
    ) -> Self:
        return cls(int.from_bytes(reverse_bits(bytes_), byteorder, signed=signed))

    def to_bytes(
        self: Self,
        length: SupportsIndex = 4,
        byteorder: Literal["little", "big"] = "big",
        *,
        signed: bool = False,
    ) -> bytes:
        return reverse_bits(super().to_bytes(length, byteorder, signed=signed))


class UnknownFlag(BitMask):
    """Unknown device flags sent by the device in the initial connection."""

    NONE = 0x00

    UNKNOWN_00 = auto()
    UNKNOWN_01 = auto()
    UNKNOWN_02 = auto()
    UNKNOWN_03 = auto()
    UNKNOWN_04 = auto()
    UNKNOWN_05 = auto()
    UNKNOWN_06 = auto()
    UNKNOWN_07 = auto()

    UNKNOWN_08 = auto()
    UNKNOWN_09 = auto()
    UNKNOWN_0A = auto()
    UNKNOWN_0B = auto()
    UNKNOWN_0C = auto()
    UNKNOWN_0D = auto()
    UNKNOWN_0E = auto()
    UNKNOWN_0F = auto()

    UNKNOWN_10 = auto()
    UNKNOWN_11 = auto()
    UNKNOWN_12 = auto()
    UNKNOWN_13 = auto()
    UNKNOWN_14 = auto()
    UNKNOWN_15 = auto()
    UNKNOWN_16 = auto()
    UNKNOWN_17 = auto()

    UNKNOWN_18 = auto()
    UNKNOWN_19 = auto()
    UNKNOWN_1A = auto()
    UNKNOWN_1B = auto()
    UNKNOWN_1C = auto()
    IS_ROOT = auto()
    """Set if the device is connecting as root for the first time."""
    UNKNOWN_1E = auto()
    UNKNOWN_1F = auto()


class Capability(BitMask):
    """Device capabilities returned by the server upon initial connection."""

    NONE = 0x00

    UNKNOWN_00 = auto()
    UNKNOWN_01 = auto()
    UNKNOWN_02 = auto()
    UNKNOWN_03 = auto()
    UNKNOWN_04 = auto()
    UNKNOWN_05 = auto()
    UNKNOWN_06 = auto()
    UNKNOWN_07 = auto()

    UNKNOWN_08 = auto()
    UNKNOWN_09 = auto()
    UNKNOWN_0A = auto()
    UNKNOWN_0B = auto()
    UNKNOWN_0C = auto()
    UNKNOWN_0D = auto()
    UNKNOWN_0E = auto()
    UNKNOWN_0F = auto()

    UNKNOWN_10 = auto()
    UNKNOWN_11 = auto()
    UNKNOWN_12 = auto()
    UNKNOWN_13 = auto()
    UNKNOWN_14 = auto()
    UNKNOWN_15 = auto()
    UNKNOWN_16 = auto()
    UNKNOWN_17 = auto()

    UNKNOWN_18 = auto()
    UNKNOWN_19 = auto()
    UNKNOWN_1A = auto()
    UNKNOWN_1B = auto()
    UNKNOWN_1C = auto()
    UNKNOWN_1D = auto()
    UNKNOWN_1E = auto()
    DUAL_CHANNEL = auto()


class Interface(IntEnum):
    """Interface used to connect to the APNs server."""

    CELLULAR = 0x00
    WIFI = 0x01


class Status(IntEnum):
    """Status codes returned by the server in response to a client's connection request."""

    OK = 0x00
    ERROR = 0x02


@dataclass
class Nonce:
    """A simple nonce used for various purposes."""

    nonce_time: datetime = field(default_factory=lambda: datetime.now().astimezone())
    random_bytes: bytes = field(default_factory=lambda: randbytes(8))
    prefix: bytes = b"\x00"

    @classmethod
    def from_bytes(cls: type[Self], data: bytes, timestamp_size: int = 8, random_bytes_length: int = 8) -> Self:
        """Deserialize a nonce from bytes."""
        stream = BytesIO(data)
        stream.read(1)  # Skip the first byte
        return cls(
            nonce_time=datetime.fromtimestamp(int.from_bytes(stream.read(timestamp_size)) / 1_000, tz=timezone.utc),
            random_bytes=stream.read(random_bytes_length),
        )

    def __bytes__(self: Self) -> bytes:
        """Serialize the nonce to bytes."""
        result = self.prefix
        result += int(self.nonce_time.timestamp() * 1_000).to_bytes(8)
        result += self.random_bytes
        return result


# -- Transformers --
NO_OP_TRANSFORMER: Final = DataTransformer(lambda data: data, lambda value: value)
PUSH_TOKEN_TRANSFORMER: Final = DataTransformer(lambda data: b64encode(data).decode(), b64decode)
UNKNOWN_FLAG_TRANSFORMER: Final = DataTransformer(
    UnknownFlag.from_bytes,
    partial(UnknownFlag.to_bytes, length=4),
)
INTERFACE_TRANSFORMER: Final = DataTransformer(Interface.from_bytes, partial(int.to_bytes, length=1))
STRING_TRANSFORMER: Final = DataTransformer(partial(bytes.decode, encoding="utf-8"), str.encode)
DER_CERTIFICATE_TRANSFORMER: Final = DataTransformer(
    load_der_x509_certificate,
    lambda certificate: certificate.public_bytes(encoding=Encoding.DER),
)
NONCE_TRANSFORMER: Final = DataTransformer(Nonce.from_bytes, bytes)
STATUS_TRANSFORMER: Final = DataTransformer(Status.from_bytes, partial(int.to_bytes, length=1))
CAPABILITIES_TRANSFORMER: Final = DataTransformer(Capability.from_bytes, partial(Capability.to_bytes, length=4))
MS_DATETIME_TRANSFORMER: Final = DataTransformer(
    lambda data: datetime.fromtimestamp(int.from_bytes(data) / 1_000, tz=timezone.utc),
    lambda value: int(value.timestamp() * 1_000).to_bytes(length=8),
)
TOPIC_TRANSFORMER: Final = DataTransformer(
    lambda data: TOPIC_HASHES.get(data, data),
    lambda value: sha1(value.encode(), usedforsecurity=False).digest() if isinstance(value, str) else value,
)


@lru_cache
def integer_transformer(size: int) -> DataTransformer[int]:
    """Create a variable-length integer transformer."""
    return DataTransformer(int.from_bytes, partial(int.to_bytes, length=size))



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""Module containing the Device class, which represents an emulated Apple device."""

from __future__ import annotations

import json
import logging
import plistlib
from base64 import b64decode, b64encode
from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import Enum, StrEnum
from hashlib import sha256
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING, Self
from urllib.parse import urlparse
from uuid import UUID, uuid4

from cryptography.hazmat.primitives.serialization import Encoding, NoEncryption, PrivateFormat, load_pem_private_key
from cryptography.x509 import load_pem_x509_certificate
from httpx import AsyncClient
from websockets import WebSocketClientProtocol, connect

from .._util.crypto import randbytes
from .._util.url import replace_url
from ..albert.activation import request_push_certificate
from ..apns.client import APNsClient
from ..ids.user import IDSUser

if TYPE_CHECKING:
    from collections.abc import Callable
    from os import PathLike

LOOKUP_URL = "https://gsa.apple.com/grandslam/GsService2/lookup"

logger = getLogger()


class OperatingSystem(StrEnum):
    """Enum containing the operating systems supported by Apple."""

    IOS = "iOS"
    MACOS = "macOS"
    WATCHOS = "watchOS"
    TVOS = "tvOS"
    WINDOWS = "Windows"

    @classmethod
    def _missing_(cls: type[Self], value: str) -> Self:
        for member in cls:
            if value.lower() == member.value.lower():
                return member
        return None


class AnisetteV3Provider(Enum):
    """Enum containing the URLs for the Anisette V3 providers."""

    SIDESTORE = "https://ani.sidestore.io/"


@dataclass
class Device:
    """Class whose objects represent an emulated Apple device."""

    name: str
    operating_system: OperatingSystem
    os_version: str
    os_build: str
    hardware_version: str
    hardware_serial: str
    save_path: PathLike | None = None
    """The folder to save the device's info folder to. If not provided, the device info will not be saved."""

    identifier: str | None = None
    user_agent: str | None = None
    client_info: str | None = None
    adi_pb: str | None = None
    machine_id: str | None = None
    one_time_password: str | None = None
    routing_info: str | None = None

    apns_client: APNsClient | None = None
    ids_users: list[IDSUser] | None = field(default_factory=list)

    provider: AnisetteV3Provider = AnisetteV3Provider.SIDESTORE

    @property
    def anisette_headers(self: Self) -> dict[str, str]:
        """The headers used in Anisette v3-privileged requests."""
        return {header: value for header, value in self._get_header_values().items() if value is not None}

    @property
    def requires_provisioning(self: Self) -> bool:
        """Whether the device requires provisioning."""
        return len(self._get_header_values()) != len(self.anisette_headers)

    @property
    def requires_push_credentials(self: Self) -> bool:
        """Whether the device requires push credentials."""
        return self.apns_client.push_key is None or self.apns_client.push_cert is None

    def __post_init__(self: Self) -> None:
        """Post-initialize the device object."""
        if self.identifier is None:
            self.identifier = b64encode(randbytes(16)).decode()

        if self.save_path is not None:
            self._save_dir = Path(self.save_path)
            logger.debug(f"Loading device info from {self._save_dir}")

        if isinstance(self._save_dir, Path):
            if self._save_dir.name != self.hardware_serial:
                self._save_dir /= self.hardware_serial

            self._users_dir = self._save_dir / "Users"
            self._users_dir.mkdir(parents=True, exist_ok=True)

            self.apns_client = APNsClient(
                os_version=self.os_version,
                os_build=self.os_build,
                hardware_version=self.hardware_version,
            )

        for user_path in self._users_dir.iterdir():
            if user_path.is_dir():
                user = IDSUser(
                    user_path.name,
                    lambda: self.anisette_headers,
                    lambda: self.apns_client.push_token,
                    lambda: self.apns_client.push_key,
                    lambda: self.apns_client.push_cert,
                )
                user.read_credentials(user_path)

                self.ids_users.append(user)

        if isinstance(self.provider, str):
            self.provider_url = urlparse(self.provider)
        else:
            self.provider_url = urlparse(self.provider.value)

        self.client: AsyncClient = AsyncClient(
            verify=False,  # noqa: S501
        )
        self.start_provisioning_url: str | None = None
        self.finish_provisioning_url: str | None = None
        self.provisioning_complete: bool = False

        self.client.headers.update(
            {
                "Content-Type": "text/x-xml-plist",
                "Accept": "*/*",
            },
        )

    async def _fetch_machine_headers(self: Self) -> None:
        """Fetch the machine headers from the Anisette provider."""
        response = await self.client.post(
            replace_url(self.provider_url, path="v3/get_headers"),
            json={
                "identifier": self.identifier,
                "adi_pb": self.adi_pb,
            },
            headers={"Content-Type": "application/json"} | self.anisette_headers,
        )

        response_data = response.json()
        logger.debug(f"Received anisette headers response: {response_data}")

        self.machine_id = response_data["X-Apple-I-MD-M"]
        self.one_time_password = response_data["X-Apple-I-MD"]
        self.routing_info = response_data["X-Apple-I-MD-RINFO"]

    async def _fetch_provisioning_urls(self: Self) -> None:
        """Fetch the provisioning URLs from Apple."""
        response = await self.client.get(LOOKUP_URL, headers=self.anisette_headers)
        response_data = plistlib.loads(response.content)
        logger.debug(f"Received fetch provisioning URLs response: {response_data}")

        urls = response_data["urls"]
        self.start_provisioning_url = urls["midStartProvisioning"]
        self.finish_provisioning_url = urls["midFinishProvisioning"]

    async def _fetch_client_info(self: Self) -> None:
        """Fetch the client info from the Anisette provider."""
        response = await self.client.get(
            replace_url(self.provider_url, path="v3/client_info"),
            headers=self.anisette_headers,
        )
        response_data = response.json()
        logger.debug(f"Received fetch client info response: {response_data}")
        self.user_agent = response_data["user_agent"]
        self.client_info = response_data["client_info"]

    async def _process_response(self: Self, ws: WebSocketClientProtocol, response: dict) -> None:
        """Process the response from the provisioning provider."""
        match response["result"]:
            case "GiveIdentifier":
                await ws.send(json.dumps({"identifier": self.identifier}))

            case "GiveStartProvisioningData":
                logger.debug(f"{self.anisette_headers = }")
                response = await self.client.post(
                    self.start_provisioning_url,
                    content=plistlib.dumps({"Header": {}, "Request": {}}),
                    headers=self.anisette_headers,
                )
                response_data = plistlib.loads(response.content)["Response"]

                logger.debug(f"Received start provisioning data: {response_data}")

                await ws.send(json.dumps({"spim": response_data["spim"]}))

            case "GiveEndProvisioningData":
                response = await self.client.post(
                    self.finish_provisioning_url,
                    content=plistlib.dumps({"Header": {}, "Request": {"cpim": response["cpim"]}}),
                    headers=self.anisette_headers,
                )
                response_data = plistlib.loads(response.content)["Response"]

                logger.debug(f"Received end provisioning data: {response_data}")

                await ws.send(json.dumps({"ptm": response_data["ptm"], "tk": response_data["tk"]}))

            case "ProvisioningSuccess":
                self.adi_pb = response["adi_pb"]
                self.provisioning_complete = True
                await self._fetch_machine_headers()
                logger.info("Provisioning complete!")

    def _get_header_values(self: Self) -> dict[str, str]:
        """Get the header values for the Anisette headers."""
        decoded_uuid = b64decode(self.identifier)
        client_time = datetime.now(tz=timezone.utc).replace(microsecond=0).isoformat()
        return {
            "User-Agent": self.user_agent,
            "X-Apple-I-MD-M": self.machine_id,
            "X-Apple-I-MD": self.one_time_password,
            "X-Apple-I-MD-RINFO": self.routing_info,
            "X-Apple-I-MD-LU": sha256(decoded_uuid).hexdigest().upper(),
            "X-Apple-I-SRL-NO": self.hardware_serial,
            "X-Apple-I-Client-Time": client_time.replace("+00:00", "") + "Z",
            "X-Apple-I-TimeZone": "UTC",
            "X-Apple-Locale": "en_US",
            "X-Mme-Client-Info": self.client_info,
            "X-Mme-Device-Id": str(UUID(bytes=decoded_uuid)).upper(),
        }

    async def start_provisioning(self: Self) -> None:
        """Start the provisioning process."""
        await self._fetch_client_info()
        logger.debug(f"{self.client.headers = }")
        await self._fetch_provisioning_urls()

        provider_url = replace_url(self.provider_url, path="v3/provisioning_session", scheme="wss")

        logger.debug(f"Starting provisioning via provider {provider_url}")

        async with connect(provider_url) as ws:
            while self.provisioning_complete is False:
                response = await ws.recv()
                result = json.loads(response)
                logger.debug(f"Received response: {result}")
                await self._process_response(ws, result)

    async def get_push_credentials(self: Self) -> None:
        """Provision the APNs client by requesting a push certificate."""
        if not self.requires_push_credentials:
            return

        self.apns_client.push_key, self.apns_client.push_cert = await request_push_certificate(
            activation_info_content={
                "ActivationRandomness": str(uuid4()).upper(),
                "ActivationState": "Unactivated",
                "DeviceClass": (
                    "MacOS" if self.operating_system is OperatingSystem.MACOS else self.operating_system.value
                ),
                "ProductType": self.hardware_version,
                "ProductVersion": self.os_version,
                "BuildVersion": self.os_build,
                "SerialNumber": self.hardware_serial,
                "UniqueDeviceID": str(UUID(bytes=b64decode(self.identifier))).upper(),
            },
        )

    def get_user(self: Self, username: str) -> IDSUser | None:
        """Get a user by their Apple ID."""
        for user in self.ids_users:
            if user.apple_id == username:
                return user
        return None

    async def add_user(
        self: Self,
        username: str,
        password_provider: Callable[[], str],
        validation_data_provider: Callable[[], str],
        two_factor_code_provider: Callable[[], str],
    ) -> IDSUser | None:
        """
        Add a user to the device.

        :param username: The Apple ID of the user.
        :param password_provider: A callable that returns the user's password.
        :param validation_data_provider: A callable that returns the user's validation data.
        :param two_factor_code_provider: A callable that returns the user's two-factor code.
        """
        user = IDSUser(
            username,
            lambda: self.anisette_headers,
            lambda: self.apns_client.push_token,
            lambda: self.apns_client.push_key,
            lambda: self.apns_client.push_cert,
        )

        if any(user.apple_id == u.apple_id and u.is_fully_authenticated for u in self.ids_users):
            msg = f"User {username} already exists on this device."
            raise ValueError(msg)

        self.ids_users.append(user)

        await user.start_authentication(password_provider())

        if user.requires_2fa and two_factor_code_provider is not None:
            await user.verify_2fa_code(two_factor_code_provider())
            if user.requires_2fa:
                logger.info("User still requires 2FA code.")
                return None

        elif user.requires_2fa:
            logger.info("User requires 2FA code.")
            return None

        user.log(logging.INFO, "Successfully authenticated user. Adding to device...")
        await user.authenticate_device()
        await user.register_device(
            self.name,
            validation_data_provider(),
            self.operating_system,
            self.hardware_version,
            self.os_version,
            self.os_build,
        )
        user.write_credentials(self._users_dir)

        return user

    @classmethod
    async def remove_user(cls: type[Self], *_args: ..., **_kwargs: ...) -> None:
        """Remove a user from the device, deregister the device from the user, and revoke their credentials."""
        msg = "User removal is not yet implemented."
        raise NotImplementedError(msg)

    async def refresh_user_handles(self: Self) -> None:
        """Refresh the user's handles, verifying their authentication status."""
        for user in self.ids_users:
            await user.fetch_handles()

    def save(self: Self) -> None:
        """Save the device's info, APNs credentials, and users to the save directory."""
        if not isinstance(self._save_dir, Path):
            return

        self._save_dir.mkdir(parents=True, exist_ok=True)

        (self._save_dir / "device_info.json").write_text(
            json.dumps(
                {
                    "name": self.name,
                    "operating_system": self.operating_system.value,
                    "os_version": self.os_version,
                    "os_build": self.os_build,
                    "hardware_version": self.hardware_version,
                    "hardware_serial": self.hardware_serial,
                    "identifier": self.identifier,
                    "user_agent": self.user_agent,
                    "client_info": self.client_info,
                    "adi_pb": self.adi_pb,
                    "machine_id": self.machine_id,
                    "one_time_password": self.one_time_password,
                    "routing_info": self.routing_info,
                },
                indent=4,
            ),
        )

        for user in self.ids_users:
            user.write_credentials(self._users_dir)

        if (push_key := self.apns_client.push_key) is not None:
            (self._save_dir / "push.key").write_bytes(
                push_key.private_bytes(Encoding.PEM, PrivateFormat.TraditionalOpenSSL, NoEncryption()),
            )

        if (push_cert := self.apns_client.push_cert) is not None:
            (self._save_dir / "push.crt").write_bytes(push_cert.public_bytes(Encoding.PEM))

        if (push_token := self.apns_client.push_token) is not None:
            (self._save_dir / "push_token.txt").write_text(push_token)

    @classmethod
    def load(cls: type[Self], path: PathLike) -> Self:
        """Load a device from the specified directory."""
        path = Path(path) if not isinstance(path, Path) else path
        data = json.loads((path / "device_info.json").read_text())

        instance = cls(
            name=data.get("name"),
            operating_system=OperatingSystem(data.get("operating_system")),
            os_version=data.get("os_version"),
            os_build=data.get("os_build"),
            hardware_version=data.get("hardware_version"),
            hardware_serial=data.get("hardware_serial"),
            user_agent=data.get("user_agent"),
            client_info=data.get("client_info"),
            identifier=data.get("identifier"),
            adi_pb=data.get("adi_pb"),
            machine_id=data.get("machine_id"),
            one_time_password=data.get("one_time_password"),
            routing_info=data.get("routing_info"),
            save_path=path,
        )

        if (push_key_path := path / "push.key").is_file():
            instance.apns_client.push_key = load_pem_private_key(push_key_path.read_bytes(), password=None)

        if (push_cert_path := path / "push.crt").is_file():
            instance.apns_client.push_cert = load_pem_x509_certificate(push_cert_path.read_bytes())

        if (push_token_path := path / "push_token.txt").is_file():
            instance.apns_client.push_token = push_token_path.read_text()

        return instance



#  Copyright (C) 2024  Cypheriel
from pathlib import Path


def get_selected_device(path: Path) -> str | None:
    selected_file = path / ".selected_device"

    if selected_file.is_file() and (selected_device := selected_file.read_text()):
        return selected_device
    return None


def set_selected_device(path: Path, serial: str) -> None:
    devices_dir = path
    if devices_dir.is_dir() and serial in (device.name for device in devices_dir.iterdir()):
        selected_file = path / ".selected_device"
        selected_file.write_text(serial)

    else:
        msg = "Failed to select device. Device not found."
        raise ValueError(msg)



#  Copyright (C) 2024  Cypheriel
from __future__ import annotations

import os
from dataclasses import dataclass
from getpass import getuser
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, ClassVar

import typer

from .._util.aio import run_async
from ..device import Device, OperatingSystem
from .util.app_dirs import USER_DATA_DIR
from .util.device_utils import fetch_device
from .util.rich_console import console
from .util.selection import get_selected_device, set_selected_device

if TYPE_CHECKING:
    from collections.abc import Generator

__ALIAS__ = "ds"


def list_devices(ctx: typer.Context, incomplete: str) -> Generator[str]:
    path = Path(ctx.params.get("path", os.environ.get("POMMEKIT_DEVICE_PATH", DeviceOptions.save_path)))

    for device_path in path.iterdir():
        if (
            device_path.is_dir()
            and (device_path / "device_info.json").is_file()
            and device_path.name.startswith(incomplete)
        ):
            yield device_path.name


SERIAL_HELP = "The serial number of the device."
SerialArgument = Annotated[
    str,
    typer.Argument(help="The serial number of the device.", show_default=False, autocompletion=list_devices),
]

app = typer.Typer()
logger = getLogger(__name__)


@dataclass(kw_only=True)
class DeviceOptions:
    save_path: ClassVar[Path] = USER_DATA_DIR / "Devices"
    selected_device: ClassVar[str | None] = None


@app.command(name="list", help="List installed devices.")
@app.command(name="ls", hidden=True)
def list_() -> None:
    if not DeviceOptions.save_path.is_dir():
        typer.echo("No devices found.", err=True)
        raise typer.Exit(1)

    found = False
    for device_path in DeviceOptions.save_path.iterdir():
        dev = fetch_device(device_path.parent, device_path.name)

        if dev is None:
            continue

        provisioned = " ([green]Provisioned[/])" if dev.requires_provisioning is False else ""
        selected = " ([blue]Selected[/])" if DeviceOptions.selected_device == dev.hardware_serial else ""
        console.print(f"{dev.name} - {dev.hardware_serial}{provisioned}{selected}")
        found = True

    if not found:
        typer.echo("No devices found.", err=True)


@app.command(help="Get information about a device.")
@app.command(name="if", hidden=True)
def info(
    serial_number: SerialArgument = DeviceOptions.selected_device,
) -> None:
    if serial_number is None:
        typer.echo("No device selected. Specify a device using `--serial-number` or the `device sel` command", err=True)
        raise typer.Exit(1)

    dev = fetch_device(DeviceOptions.save_path, serial_number)
    typer.echo(
        f"Device: {dev.name}\n"
        f"Serial Number: {dev.hardware_serial}\n"
        f"UID: {dev.identifier}\n"
        f"Product Type: {dev.hardware_version}\n"
        f"Product Version: {dev.os_version}\n"
        f"Build Version: {dev.os_build}\n"
        f"Device Class: {dev.operating_system.value}\n",
    )


@app.command(help="Create a new device.")
@app.command(name="create", hidden=True)
@app.command(name="new", hidden=True)
@run_async
async def add(
    device_name: Annotated[
        str,
        typer.Option(
            prompt=True,
            help="The name of the device.",
        ),
    ] = f"{getuser()}'s PC",
    serial_number: Annotated[
        str,
        typer.Option(
            prompt=True,
            help=SERIAL_HELP,
            autocompletion=list_devices,
        ),
    ] = "WindowSerial",
    operating_system: Annotated[
        OperatingSystem,
        typer.Option(
            prompt=True,
            help="The operating system of the device.",
        ),
    ] = OperatingSystem.WINDOWS,
    os_version: Annotated[
        str,
        typer.Option(
            prompt="OS version",
            help="The version of the operating system.",
        ),
    ] = "10.6.4",
    os_build: Annotated[
        str,
        typer.Option(
            prompt="OS build",
            help="The build of the operating system.",
        ),
    ] = "10.6.4",
    hardware_version: Annotated[
        str,
        typer.Option(
            prompt=True,
            help="The hardware version of the device.",
        ),
    ] = "windows1,1",
    *,
    skip_provisioning: Annotated[
        bool,
        typer.Option(
            "--skip-provisioning",
            "-s",
            help="Skip Anisette provisioning.",
        ),
    ] = False,
) -> None:
    if fetch_device(DeviceOptions.save_path, serial_number) is not None:
        typer.confirm("Device already exists... Overwrite?", abort=True)

    dev = Device(
        name=device_name,
        hardware_serial=serial_number,
        operating_system=operating_system,
        os_version=os_version,
        os_build=os_build,
        hardware_version=hardware_version,
        save_path=DeviceOptions.save_path,
    )

    dev.save()

    if dev.requires_provisioning is False or skip_provisioning is True:
        typer.echo("Skipping Anisette provisioning.")
        return

    typer.echo("Starting Anisette provisioning...")
    await dev.start_provisioning()

    typer.echo("Fetching push credentials from Albert...")
    if dev.requires_push_credentials:
        await dev.get_push_credentials()

    dev.save()


@app.command(help="Remove a device.")
@app.command(name="rm", hidden=True)
@app.command(name="delete", hidden=True)
@run_async
async def remove() -> None: ...


@app.command(help="Provision a device through Anisette.")
@app.command(name="prov", hidden=True)
@run_async
async def provision(
    serial_number: SerialArgument,
    *,
    force: Annotated[
        bool,
        typer.Option(
            "--force",
            "-f",
            help="Force a re-provision of the device.",
        ),
    ] = False,
) -> None:
    dev = fetch_device(DeviceOptions.save_path, serial_number)

    if force is True:
        logger.warning(
            "Re-provisioning the Anisette data and/or push credentials may cause issues with authenticated users!",
        )

    if dev.requires_provisioning is True or force is True:
        await dev.start_provisioning()

    if dev.requires_push_credentials is True or force is True:
        await dev.get_push_credentials()

    else:
        logger.error("Device is already provisioned. Use --force to re-provision.")
        raise typer.Abort

    dev.save()


@app.command(help="Select this device for use with other PommeKit commands.")
@app.command(name="sel", hidden=True)
def select(serial_number: SerialArgument = None) -> None:
    if serial_number is None:
        if DeviceOptions.selected_device is None:
            raise typer.Exit

        typer.echo(DeviceOptions.selected_device)
        raise typer.Exit

    set_selected_device(DeviceOptions.save_path, serial_number)


@app.callback(no_args_is_help=True, hidden=False)
def device(
    path: Annotated[
        Path | None,
        typer.Option(
            "--path",
            "-p",
            envvar="POMMEKIT_DEVICE_PATH",
            help="The path to save/load device data to.",
        ),
    ] = DeviceOptions.save_path,
) -> None:
    DeviceOptions.save_path = path
    DeviceOptions.selected_device = get_selected_device(path)

    logger.debug(f"Device path set to '{DeviceOptions.save_path}'.")
    logger.debug(f"Selected device set to '{DeviceOptions.selected_device}'.")



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel

"""Module for generating a `CertificateSigningRequest` used for the request to Albert."""

from uuid import uuid4

from cryptography import x509
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.x509 import (
    CertificateSigningRequest,
    CertificateSigningRequestBuilder,
)
from cryptography.x509.oid import NameOID


def generate_device_csr(private_key: RSAPrivateKey) -> CertificateSigningRequest:
    """Generate a `CertificateSigningRequest` used for the request to Albert."""
    return CertificateSigningRequestBuilder(
        subject_name=x509.Name(
            [
                x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
                x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "CA"),
                x509.NameAttribute(NameOID.LOCALITY_NAME, "Cupertino"),
                x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Apple Inc."),
                x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, "iPhone"),
                x509.NameAttribute(NameOID.COMMON_NAME, str(uuid4())),
            ],
        ),
    ).sign(private_key, SHA256())



"""
Copyright (c) 2024  Cypheriel.

Secure Remote Password protocol implementation.

This module provides an implementation of the Secure Remote Password protocol, as defined in RFC 5054.
For now, only the client-side implementation is provided.
This implementation is designed to be compatible with the Apple SRP implementation, as used in the GrandSlam framework.

See:
  - https://datatracker.ietf.org/doc/html/rfc5054
"""

from __future__ import annotations

from functools import lru_cache
from hashlib import sha256
from importlib import resources
from typing import TYPE_CHECKING, Self

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import dh

from .._util.crypto import randbytes

if TYPE_CHECKING:
    from os import PathLike


@lru_cache
def _byte_length(value: int) -> int:
    """
    Calculate the byte length of an integer.

    :param value: The integer to calculate the byte length of.
    :return: The length of the integer in bytes.

    >>> _byte_length(255)
    1

    >>> _byte_length(256)
    2
    """
    return (value.bit_length() + 7) // 8


@lru_cache
def _to_bytes(value: int) -> bytes:
    r"""
    Convert an integer to a dynamically-sized bytes object.

    :param value: The integer to convert.
    :return: The bytes object.

    >>> _to_bytes(255)
    b'\xff'

    >>> _to_bytes(256)
    b'\x01\x00'
    """
    return value.to_bytes(_byte_length(value))


def _generate_safe_prime(generator: int = 2, key_size: int = 2048) -> tuple[int, int]:
    """
    Generate a safe prime and generator using cryptography's DH module.

    :return: The safe prime and generator.
    """
    dh_params = dh.generate_parameters(generator, key_size)
    return dh_params.parameter_numbers().p, dh_params.parameter_numbers().g


def _load_safe_prime(path: PathLike | None = None) -> tuple[int, int]:
    """
    Load the safe prime and generator from a DH parameters PEM file.

    By default, the parameters are loaded from the package resources.

    :param path: The path to the DHParams PEM file.
    :return: The safe prime and generator.
    """
    if path is None:
        path = resources.files(__package__) / "params.pem"

    with path.open("rb") as file:
        dh_params = serialization.load_pem_parameters(file.read())

    return dh_params.parameter_numbers().p, dh_params.parameter_numbers().g


@lru_cache
def _hash(*args: int | bytes, width: int | None = None, include_headers: bool = True) -> int:
    """
    Hash the provided arguments using SHA-256.

    In compliance with RFC 5054, the hash (by default) includes length headers for each argument.

    :param args: The arguments to hash.
    :param width: The width of the hash.
    :param include_headers: Whether to include length headers.
    :return: The hashed value, as an integer.
    """
    hash_data = sha256()

    for arg in args:
        if not isinstance(arg, (int, bytes)):
            msg = f"Expected int | bytes, got {type(arg)}."
            raise TypeError(msg)

        arg_data = _to_bytes(arg) if isinstance(arg, int) else arg

        if include_headers is True and width is not None:
            hash_data.update(bytes(width - len(arg_data)))

        hash_data.update(arg_data)

    return int.from_bytes(hash_data.digest())


class SRPUser:
    """
    Mostly RFC 5054-compatible (by default) implementation of the Secure Remote Password protocol.

    This class is used to generate the client proof, and verify the server proof.
    """

    def __init__(
        self: Self,
        username: str,
        safe_prime: int | None = None,
        generator: int | None = None,
        private_ephemeral: int | None = None,
        *,
        generate_safe_prime: bool = False,
    ) -> None:
        """Initialize the SRP user."""
        self.username = username.encode()  # I
        """
        The user's identifying username.
        This value is provided by the user, and sent to the server, along with the user's public ephemeral value.

        `I`
        """

        _prime_args = (safe_prime, generator)
        _any_prime_args = any(_prime_args)
        if _any_prime_args and not all(_prime_args):
            msg = "All or none of safe_prime, generator must be provided."
            raise ValueError(msg)

        if not _any_prime_args and generate_safe_prime is False:
            safe_prime, generator = _load_safe_prime()

        elif generate_safe_prime is True:
            safe_prime, generator = _generate_safe_prime()

        self.safe_prime = safe_prime
        """
        The safe prime number.

        `N = 2q+1`, where q is prime
        """

        self.generator = generator
        """
        The generator number.

        `g`
        """

        self.multiplier = _hash(self.safe_prime, self.generator, width=_byte_length(self.safe_prime))
        """
        SRP-6a multiplier parameter

        `k = H(N, g)`
        """

        self.private_ephemeral: int = private_ephemeral or int.from_bytes(randbytes(32))
        """
        User's private ephemeral value, a random number.

        `a = random(32)`
        """

        self.public_ephemeral: bytes = _to_bytes(pow(self.generator, self.private_ephemeral, self.safe_prime))
        """
        User's public ephemeral value. This value is sent to the server, along with the user's username.

        `A = g^a % N`
        """

        self.salt: bytes | None = None
        """
        Salt received from the server.

        `s`
        """

        self.server_public_ephemeral: int | None = None
        """
        The server's public ephemeral value.

        `B`
        """

        self.scrambling_parameter: int | None = None
        """
        The scrambling parameter.

        `u = H(A, B)`
        """

        self.private_key: int | None = None
        """
        The user's private key.

        `x = H(s, H(I, ":", P))`
        """

        self.verifier: int | None = None
        """
        The server's password verifier.

        `v = g^x % N`
        """

        self.session_key: bytes | None = None
        """
        The user's session key.

        `S = (B - k * g^x)^(a + u * x) % N`
        """

        self.shared_key: bytes | None = None
        """
        The shared key.

        `K = H(S)`
        """

        self.client_proof: bytes | None = None
        """
        The client proof. This value is sent to the server to verify the user.

        `M(User) = H(H(N) XOR H(g), H(I), s, A, B, K)`
        """

        self.server_proof: bytes | None = None
        """
        The server proof, this value is used to verify the server.

        `M(Server) = H(A, M, K)`
        """

    def _xor_prime_and_generator_bytes(self: Self, *, include_padding: bool = True) -> bytes:
        """
        XOR each byte of the hashed safe prime and generator.

        :param include_padding: Whether to include padding in the generator bytes.
        :return: The XORed bytes.
        """
        generator = self.generator.to_bytes(_byte_length(self.safe_prime))
        if include_padding is False:
            generator = generator.lstrip(b"\x00")

        prime_hashed = sha256(_to_bytes(self.safe_prime)).digest()
        generator_hashed = sha256(generator).digest()

        return b"".join((p_byte ^ g_byte).to_bytes() for p_byte, g_byte in zip(prime_hashed, generator_hashed))

    def _calculate_client_proof(self: Self) -> bytes:
        """
        Calculate the client proof.

        :return: The client proof.
        """
        return sha256(
            self._xor_prime_and_generator_bytes()
            + sha256(self.username).digest()
            + self.salt
            + self.public_ephemeral
            + _to_bytes(self.server_public_ephemeral)
            + self.shared_key,
        ).digest()

    def _generate_private_key(self: Self, password: bytes, salt: int | bytes, *, include_username: bool = False) -> int:
        """
        Generate the private key for the user.

        :param password: The user's password.
        :param salt: The salt provided by the server.
        :param include_username: Whether to include the username in the hash.
        :return:
        """
        username = self.username if include_username is True else b""

        return _hash(salt, _hash(username, b":", password, include_headers=False))

    def process_challenge(
        self: Self,
        password: str | bytes,
        salt: int | bytes,
        server_public_ephemeral: int | bytes,
    ) -> bytes:
        """
        Process the challenge from the server and generate the client proof.

        :param password: The user's password.
        :param salt: The salt provided by the server.
        :param server_public_ephemeral: The server's public ephemeral.
        :return: The client proof, also stored under `self.client_proof`.
        """
        if not isinstance(password, bytes):
            password = password.encode()

        if isinstance(server_public_ephemeral, bytes):
            server_public_ephemeral = int.from_bytes(server_public_ephemeral)

        if isinstance(salt, int):
            salt = _to_bytes(salt)

        if server_public_ephemeral % self.safe_prime == 0:
            msg = "Server public ephemeral is invalid."
            raise ValueError(msg)

        self.salt = salt
        self.server_public_ephemeral = server_public_ephemeral

        self.scrambling_parameter = _hash(
            self.public_ephemeral,
            self.server_public_ephemeral,
            width=_byte_length(self.safe_prime),
        )
        self.private_key = self._generate_private_key(password, salt)

        self.verifier = pow(self.generator, self.private_key, self.safe_prime)

        self.session_key = _to_bytes(
            pow(
                (self.server_public_ephemeral - self.multiplier * self.verifier),
                (self.private_ephemeral + self.scrambling_parameter * self.private_key),
                self.safe_prime,
            ),
        )

        self.shared_key = sha256(self.session_key).digest()

        self.client_proof = self._calculate_client_proof()

        self.server_proof = sha256(self.public_ephemeral + self.client_proof + self.shared_key).digest()

        return self.client_proof

    def verify_session(self: Self, server_proof: bytes) -> bool:
        """
        Verify the server proof.

        :param server_proof: The server proof.
        :return: Whether the server proof is valid.
        """
        return server_proof == self.server_proof



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel

"""Apple Push Notification service (APNs) credential provisioning service — "Albert"."""



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class ExponentialBackoff:
    base_delay: float = 1.0
    max_delay: float = 60.0
    max_retries: int = 10

    _retries: int = 0
    _delay: float = 0.0

    def next(self: ExponentialBackoff) -> float:
        """
        Return the next delay in seconds for the exponential backoff.

        >>> backoff = ExponentialBackoff(base_delay=4.0)
        >>> backoff.next()
        4.0
        >>> backoff.next()
        8.0
        >>> backoff.next()
        16.0

        :return: The next delay in seconds.
        """
        if self._retries >= self.max_retries:
            msg = "Max retries exceeded"
            raise StopIteration(msg)

        self._delay = min(self.base_delay * 2**self._retries, self.max_delay)
        self._retries += 1
        return self._delay



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel

"""General utilities designed to facilitate connections with the Apple Push Notification service (APNs)."""



#  Copyright (C) 2024  Cypheriel
from pathlib import Path

from pommekit._cli.util.app_dirs import USER_DATA_DIR
from pommekit.device import Device


def fetch_device(path: Path, serial_number: str) -> Device | None:
    if not path and not serial_number:
        msg = "Either a path or serial number must be provided."
        raise ValueError(msg)

    path = path / serial_number or USER_DATA_DIR / "Devices" / serial_number
    if not path.is_dir():
        return None

    return Device.load(path)



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel

import asyncio
from collections.abc import Hashable
from enum import Enum, auto
from logging import getLogger
from typing import Annotated, Awaitable, Callable

logger = getLogger(__name__)


def _verify_callback(func: Callable[[...], Awaitable[None]]) -> None:
    if not callable(func):
        msg = f"Listener must be a callable, got {type(func).__name__}!"
        raise TypeError(msg)

    if not asyncio.iscoroutinefunction(func):
        msg = f"Listener must be a coroutine function, got {type(func).__name__}!"
        raise TypeError(msg)


def _get_identifier_name(identifier: Annotated[Hashable, "Identifier"]) -> str:
    if isinstance(identifier, (bytes, str)):
        identifier_name = f"{identifier!r}"
    elif isinstance(identifier, Enum):
        identifier_name = f"{identifier.name}"
    elif hasattr(identifier, "__qualname__"):
        identifier_name = f"{identifier.__qualname__}"
    else:
        identifier_name = f"{identifier}"

    return identifier_name


class EventType(Enum):
    COMMAND = auto()
    TOPIC = auto()
    CONNECT = auto()
    OTHER = auto()


class EventListener:
    def __init__(self: "EventListener") -> None:
        self._callback_map: dict[
            tuple[EventType, Annotated[Hashable, "Identifier"]],
            list[Callable[[...], Awaitable[None]]],
        ] = {}

    def _register_event_listener(
        self: "EventListener",
        event_type: EventType,
        identifier: Annotated[Hashable, "Identifier"],
        *,
        internal: bool = True,
    ) -> Callable:
        def decorator(func: Callable) -> Callable:
            _verify_callback(func)
            internal_s = "internal " if internal else ""
            identifier_name = _get_identifier_name(identifier)
            logger.debug(
                f"Registering {internal_s}listener {func.__name__}() for {event_type.name} event {identifier_name}.",
            )

            if (event_type, identifier) in self._callback_map:
                self._callback_map[event_type, identifier].append(func)
            else:
                self._callback_map[event_type, identifier] = [func]

            return func

        return decorator

    async def _trigger_event(
        self: "EventListener",
        event_type: EventType,
        identifier: Annotated[Hashable, "Identifier"],
        *args: ...,
        **kwargs: ...,
    ) -> None:
        logger.debug(f"Triggering {event_type.name} event {_get_identifier_name(identifier)}.")

        listeners = self._callback_map.get((event_type, identifier))

        if listeners:
            await asyncio.gather(*(listener(*args, **kwargs) for listener in listeners))



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""Module containing functions to fetch Apple's APNs and IDS bags."""

from __future__ import annotations

import asyncio
import plistlib
from functools import lru_cache
from logging import getLogger
from typing import Final

from httpx import AsyncClient

logger = getLogger(__name__)
client = AsyncClient(
    verify=False,  # noqa: S501
)


class BagFetchError(Exception):
    """Exception raised when a bag fetch fails."""


@lru_cache
async def get_apns_bag() -> dict[str, str]:
    """Fetch Apple's APNs bag."""
    response = await client.get(
        "https://init.push.apple.com/bag",
    )

    if not response.is_success:
        msg = f"Failed to fetch APNs bag! Status: {response.status_code}"
        raise BagFetchError(msg)

    return plistlib.loads(response.content)


@lru_cache
async def get_ids_bag() -> dict[str, str]:
    """Fetch Apple's IDS bag."""
    response = await client.get("https://init.ess.apple.com/WebObjects/VCInit.woa/wa/getBag?ix=3")

    if not response.is_success:
        msg = "Failed to fetch IDS bag!"
        raise BagFetchError(msg)

    return plistlib.loads(plistlib.loads(response.content)["bag"])


apns_bag: Final = {}
ids_bag: Final = {}


async def _fetch_bags() -> None:
    apns_bag.update(await get_apns_bag())
    ids_bag.update(await get_ids_bag())


asyncio.run(_fetch_bags())



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel

from __future__ import annotations

from urllib.parse import ParseResult, urlparse


def replace_url(url: str | ParseResult, scheme: str | None = None, path: str | None = None) -> str:
    """
    Replace the scheme and/or path of a URL.

    >>> replace_url("https://example.com/path/to/resource", scheme="wss", path="/new/path")
    'wss://example.com/new/path'

    :param url: The URL to replace the scheme and/or path of.
    :param scheme: The new scheme to use. If `None`, the scheme will not be changed.
    :param path: The new path to use. If `None`, the path will not be changed.
    :return: The URL with the new scheme and/or path.
    """
    parsed = url if isinstance(url, ParseResult) else urlparse(url)
    if scheme is not None:
        parsed = parsed._replace(scheme=scheme)
    if path is not None:
        parsed = parsed._replace(path=path)

    return parsed.geturl()



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""Submodule containing known APNs commands and their respective packet fields as well as a command registry."""

from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass, field
from datetime import datetime
from enum import Flag, auto
from logging import getLogger
from typing import Annotated, Collection, Final, Literal, TypeAlias, TypeVar

from cryptography.x509 import Certificate

from ..protocol.packet import APNsCommand, Item
from ..protocol.transformers import (
    CAPABILITIES_TRANSFORMER,
    DER_CERTIFICATE_TRANSFORMER,
    INTERFACE_TRANSFORMER,
    MS_DATETIME_TRANSFORMER,
    NO_OP_TRANSFORMER,
    NONCE_TRANSFORMER,
    PUSH_TOKEN_TRANSFORMER,
    STATUS_TRANSFORMER,
    STRING_TRANSFORMER,
    TOPIC_TRANSFORMER,
    UNKNOWN_FLAG_TRANSFORMER,
    Capability,
    Interface,
    Nonce,
    Status,
    UnknownFlag,
    integer_transformer,
)


@dataclass
class UnknownCommand(APNsCommand):
    """Dummy command class to encapsulate unknown commands."""


class Direction(Flag):
    """Enumeration of the possible directions a command can be sent in."""

    CLIENT_BOUND = auto()
    SERVER_BOUND = auto()
    BIDIRECTIONAL = CLIENT_BOUND | SERVER_BOUND


COMMAND_MAP: Final[defaultdict[(int, Direction), type[APNsCommand]]] = defaultdict()
"""Map of command IDs to their packet classes."""

COMMAND_MAP.default_factory = lambda: register_command(-0x01, Direction.BIDIRECTIONAL)(UnknownCommand)

_T_co = TypeVar("_T_co", covariant=True, bound=type[APNsCommand])
_FIELDS: Final = "__dataclass_fields__"

Topic: TypeAlias = str | bytes

logger = getLogger(__name__)


def register_command(command_id: int, direction: Direction) -> Callable[[_T_co], _T_co]:
    """Register an ``APNsCommand`` subclass with the command registry."""

    def wrapper(cls: _T_co) -> _T_co:
        """Set the command's command ID and register it with the command registry."""
        cls.command_id = command_id

        for d in direction:
            logger.debug(f"Registering command {cls.__name__} (0x{command_id:02X}) for {d.name} direction.")
            COMMAND_MAP[(command_id, d)] = cls

        return cls

    return wrapper


command = dataclass(repr=False)


@register_command(0x07, Direction.SERVER_BOUND)
@dataclass(repr=False)
class ConnectCommand(APNsCommand):
    """Command sent by the client to establish a connection with the APNs server."""

    push_token: Annotated[str | None, Item(0x01, PUSH_TOKEN_TRANSFORMER)]

    certificate: Annotated[Certificate | None, Item(0x0C, DER_CERTIFICATE_TRANSFORMER)]
    nonce: Annotated[Nonce | None, Item(0x0D, NONCE_TRANSFORMER)]
    signature: Annotated[bytes | None, Item(0x0E)]

    state: Annotated[Literal[b"\x01"], Item(0x02)] = b"\x01"
    flags: Annotated[UnknownFlag, Item(0x05, UNKNOWN_FLAG_TRANSFORMER)] = UnknownFlag.IS_ROOT
    interface: Annotated[Interface, Item(0x06, INTERFACE_TRANSFORMER)] = Interface.WIFI
    carrier: Annotated[str, Item(0x08, STRING_TRANSFORMER)] = "WiFi"
    os_version: Annotated[str, Item(0x09, STRING_TRANSFORMER)] = "10.6.4"
    os_build: Annotated[str, Item(0x0A, STRING_TRANSFORMER)] = "10.6.4"
    hardware_version: Annotated[str, Item(0x0B, STRING_TRANSFORMER)] = "windows1,1"
    protocol_version__: Annotated[int | None, Item(0x10, integer_transformer(2))] = 2
    redirect_count: Annotated[int, Item(0x11, integer_transformer(2))] = 0
    dns_resolve_time_ms: Annotated[int | None, Item(0x13, integer_transformer(2))] = None
    tls_handshake_time_ms: Annotated[int | None, Item(0x14, integer_transformer(2))] = None


@register_command(0x08, Direction.CLIENT_BOUND)
@dataclass(repr=False)
class ConnectResponseCommand(APNsCommand):
    """Command sent by the server in response to a client's connection request."""

    status: Annotated[Status, Item(0x01, STATUS_TRANSFORMER)]

    server_metadata__: Annotated[bytes | None, Item(0x02)]
    push_token: Annotated[str | None, Item(0x03, PUSH_TOKEN_TRANSFORMER)]
    max_message_size: Annotated[int | None, Item(0x04, integer_transformer(4))]
    protocol_version__: Annotated[int | None, Item(0x05, integer_transformer(4))]
    capabilities: Annotated[Capability | None, Item(0x06, CAPABILITIES_TRANSFORMER)]
    bad_nonce_time__: Annotated[datetime | None, Item(0x07, MS_DATETIME_TRANSFORMER)]
    large_message_size: Annotated[int | None, Item(0x08, integer_transformer(4))]
    server_time: Annotated[datetime | None, Item(0x0A, MS_DATETIME_TRANSFORMER)]
    geo_region: Annotated[str | None, Item(0x0B, STRING_TRANSFORMER)]
    unknown_timestamp__: Annotated[datetime | None, Item(0x0C, MS_DATETIME_TRANSFORMER)]


@register_command(0x09, Direction.SERVER_BOUND)
@dataclass(repr=False)
class FilterTopicsCommand(APNsCommand):
    """Command sent by the client to filter incoming push notifications based on topics."""

    push_token: Annotated[str, Item(0x01, PUSH_TOKEN_TRANSFORMER)]
    enabled_topics: Annotated[Collection[Topic] | None, Item(0x02, TOPIC_TRANSFORMER)] = field(
        default_factory=frozenset,
    )
    disabled_topics: Annotated[Collection[Topic] | None, Item(0x03, TOPIC_TRANSFORMER)] = field(
        default_factory=frozenset,
    )
    opportunity_topics: Annotated[Collection[Topic] | None, Item(0x04, TOPIC_TRANSFORMER)] = field(
        default_factory=frozenset,
    )
    paused_topics: Annotated[Collection[Topic] | None, Item(0x05, TOPIC_TRANSFORMER)] = field(default_factory=frozenset)


@register_command(0x0A, Direction.CLIENT_BOUND)
@dataclass(repr=False)
class ClientBoundPushNotificationCommand(APNsCommand):
    """Command sent by the server to deliver a push notification to a client."""

    push_token: Annotated[str, Item(0x01, PUSH_TOKEN_TRANSFORMER)]
    topic: Annotated[Topic, Item(0x02, TOPIC_TRANSFORMER)]
    payload: Annotated[bytes, Item(0x03, NO_OP_TRANSFORMER)]
    message_id: Annotated[int, Item(0x04, integer_transformer(4))]
    unknown__: Annotated[Literal[b"\x00"], Item(0x07, NO_OP_TRANSFORMER)] = b"\x00"


@register_command(0x0A, Direction.SERVER_BOUND)
@dataclass(repr=False)
class ServerBoundPushNotificationCommand(APNsCommand):
    """Command sent by the client to deliver a push notification to the server."""

    topic: Annotated[Topic, Item(0x01, TOPIC_TRANSFORMER)]
    push_token: Annotated[str, Item(0x02, PUSH_TOKEN_TRANSFORMER)]
    payload: Annotated[bytes, Item(0x03, NO_OP_TRANSFORMER)]
    message_id: Annotated[int, Item(0x04, integer_transformer(4))]
    unknown__: Annotated[Literal[b"\x00"], Item(0x07, NO_OP_TRANSFORMER)] = b"\x00"


@register_command(0x0B, Direction.BIDIRECTIONAL)
@dataclass(repr=False)
class PushNotificationAckCommand(APNsCommand):
    """Command sent by the recipient to acknowledge the receipt of a push notification."""

    message_id: Annotated[int, Item(0x04, integer_transformer(4))]
    status: Annotated[Status, Item(0x08, STATUS_TRANSFORMER)]


@register_command(0x0C, Direction.SERVER_BOUND)
@dataclass(repr=False)
class KeepAliveCommand(APNsCommand):
    """Command sent by the client to request confirmation that the connection is still alive."""

    connection_method: Annotated[str | None, Item(0x01, STRING_TRANSFORMER)]
    os_version: Annotated[str | None, Item(0x02, STRING_TRANSFORMER)]
    os_build: Annotated[str | None, Item(0x03, STRING_TRANSFORMER)]
    hardware_version: Annotated[str | None, Item(0x04, STRING_TRANSFORMER)]


@register_command(0x0D, Direction.CLIENT_BOUND)
@dataclass(repr=False)
class KeepAliveConfirmationCommand(APNsCommand):
    """Command sent by the server to confirm that the connection is still alive."""


@register_command(0x0E, Direction.CLIENT_BOUND)
@dataclass(repr=False)
class NoStorageCommand(APNsCommand):
    """Command sent by the server to inform the client that it has no storage available."""

    push_token: Annotated[str, Item(0x03, PUSH_TOKEN_TRANSFORMER)]


@register_command(0x0F, Direction.BIDIRECTIONAL)
@dataclass(repr=False)
class FlushCommand(APNsCommand):
    """Command sent by either the client or the server to flush the stream."""



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""Module containing the IDSUser class."""

from __future__ import annotations

import asyncio
import json
import logging
import plistlib
from asyncio import Event
from base64 import b64decode, b64encode
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Final, Literal, Self
from uuid import uuid4

from cryptography.hazmat.primitives.asymmetric import ec, rsa
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.hashes import SHA1, SHA256
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    load_pem_private_key,
)
from cryptography.x509 import (
    Certificate,
    CertificateSigningRequestBuilder,
    Name,
    NameAttribute,
    NameOID,
    load_der_x509_certificate,
    load_pem_x509_certificate,
)
from httpx import AsyncClient

from .._util.crypto import b64encoded_der, construct_identity_key, randbytes
from .._util.exponential_backoff import ExponentialBackoff
from ..apns.protocol.transformers import PUSH_TOKEN_TRANSFORMER, Nonce
from ..bags import ids_bag
from ..gsa import AUStatus, GSAClient
from ..status_codes import AppleStatusCode

if TYPE_CHECKING:
    from collections.abc import Callable
    from os import PathLike

    from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey
    from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

AUTHENTICATE_USER_KEY: Final = "vc-profile-authenticate"
AUTHENTICATE_USER_URL: Final = ids_bag[AUTHENTICATE_USER_KEY]

AUTHENTICATE_DEVICE_KEY: Final = "id-authenticate-ds-id"
AUTHENTICATE_DEVICE_URL: Final = ids_bag[AUTHENTICATE_DEVICE_KEY]

GET_HANDLES_KEY: Final = "id-get-handles"
GET_HANDLES_URL: Final = ids_bag[GET_HANDLES_KEY]

REGISTER_DEVICE_KEY: Final = "id-register"
REGISTER_DEVICE_URL: Final[str] = ids_bag[REGISTER_DEVICE_KEY]


logger = logging.getLogger(__name__)


class IDSError(Exception):
    """Generic exception for IDS-related errors."""


class IDSTwoFactorRequiredError(IDSError):
    """Exception raised when two-factor authentication is required but not given."""


class IDSTwoFactorInvalidError(IDSError):
    """Exception raised when the two-factor authentication code is invalid."""


def _to_length_value(value: bytes, length_size: int = 4) -> bytes:
    """Prepend the length of the value to the value."""
    return len(value).to_bytes(length_size) + value


class IDSUser:
    """Class representing a single Apple ID user."""

    @property
    def is_authenticated(self: Self) -> bool:
        """Whether the user is authenticated."""
        authenticated = self.on_authenticated_event.is_set()
        if not authenticated:
            authenticated = self.auth_token is not None and self.profile_id is not None

        return authenticated

    @property
    def is_device_authenticated(self: Self) -> bool:
        """Whether the user is device-authenticated."""
        device_authenticated = self.on_device_authenticated_event.is_set()
        if not device_authenticated:
            device_authenticated = self.private_key is not None and self.certificate is not None

        return device_authenticated

    @property
    def is_fully_authenticated(self: Self) -> bool:
        """Whether the user is fully authenticated and handles are available."""
        return len(self.handles) > 0

    @property
    def is_device_registered(self: Self) -> bool:
        """Whether the device is registered under the user."""
        return (
            self.device_certificate is not None
            and datetime.now().astimezone() < self.device_certificate.not_valid_after_utc.astimezone()
        )

    def __init__(
        self: Self,
        apple_id: str,
        anisette_provider: Callable[[], dict[str, str]],
        push_token_provider: Callable[[], str] | None = None,
        push_key_provider: Callable[[], RSAPrivateKey] | None = None,
        push_cert_provider: Callable[[], Certificate] | None = None,
    ) -> None:
        """Initialize a new instance of the IDSUser class."""
        self.apple_id = apple_id
        self._anisette_provider = anisette_provider

        self.push_key_provider = push_key_provider
        self.push_cert_provider = push_cert_provider
        self.push_token_provider = lambda: PUSH_TOKEN_TRANSFORMER.serialize(push_token_provider())

        self.password: str | None = None
        self.private_key: RSAPrivateKey | None = None
        self.certificate: Certificate | None = None
        self.profile_id: str | None = None
        self.auth_token: str | None = None
        self.signing_key: EllipticCurvePrivateKey | None = None
        self.encryption_key: RSAPrivateKey | None = None
        self.identity_key: bytes | None = None
        self.device_certificate: Certificate | None = None

        self.handles: list[dict[Literal["uri"], str]] = []

        self.requires_2fa: bool = False

        self.on_authenticated_event = Event()
        self.on_device_authenticated_event = Event()
        self.on_device_registered_event = Event()

        self._credential_file_names = {
            "auth_token": "auth_token.txt",
            "profile_id": "profile_id.txt",
            "private_key": "user.key",
            "certificate": "user.crt",
            "signing_key": "signing.key",
            "encryption_key": "encryption.key",
            "identity_key": "identity.key",
            "device_certificate": "device.crt",
        }

        self._client = AsyncClient(
            verify=False,  # noqa: S501
        )
        self._gsa_client = GSAClient(self._anisette_provider)
        self._2fa_code: str | None = None

        self._client.headers.update(
            {
                "x-protocol-version": "1630",
            },
        )

        self._auth_backoff = ExponentialBackoff(base_delay=4, max_retries=12)

    def _ensure_random_credentials(self: Self) -> None:
        """Ensure that random credentials are generated if they are not already present."""
        if self.private_key is None:
            self.private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

        if self.signing_key is None:
            self.signing_key = ec.generate_private_key(ec.SECP256R1())

        if self.encryption_key is None:
            self.encryption_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

        if self.identity_key is None:
            self.identity_key = construct_identity_key(self.signing_key.public_key(), self.encryption_key.public_key())

    def _generate_payload(
        self: Self,
        bag_key: str = "",
        query_string: str = "",
        payload: bytes = b"",
    ) -> tuple[bytes, bytes]:
        """Generate a payload for authentication."""
        nonce = bytes(Nonce(prefix=b"\x01"))

        return (
            b64encode(nonce),
            nonce
            + _to_length_value(bag_key.encode())
            + _to_length_value(query_string.encode())
            + _to_length_value(payload)
            + _to_length_value(self.push_token_provider()),
        )

    def _generate_signed_payload(
        self: Self,
        private_key: RSAPrivateKey,
        bag_key: str = "",
        query_string: str = "",
        payload: bytes = b"",
    ) -> tuple[bytes, bytes]:
        nonce, payload = self._generate_payload(bag_key, query_string, payload)
        signed_payload = b64encode(b"\x01\x01" + private_key.sign(payload, PKCS1v15(), SHA1()))  # noqa: S303
        return nonce, signed_payload

    def _generate_auth_headers(
        self: Self,
        bag_key: str = "",
        payload: bytes = b"",
        auth_suffix_number: int | None = None,
    ) -> dict[str, str]:
        push_nonce, push_sig = self._generate_signed_payload(self.push_key_provider(), bag_key=bag_key, payload=payload)
        auth_nonce, auth_sig = self._generate_signed_payload(self.private_key, bag_key=bag_key, payload=payload)

        auth_suffix = f"-{auth_suffix_number}" if auth_suffix_number is not None else ""
        return {
            "x-push-sig": push_sig,
            "x-push-nonce": push_nonce,
            "x-push-cert": b64encoded_der(self.push_cert_provider()),
            "x-push-token": PUSH_TOKEN_TRANSFORMER.deserialize(self.push_token_provider()),
            "x-auth-user-id": self.profile_id,
            f"x-auth-sig{auth_suffix}": auth_sig,
            f"x-auth-nonce{auth_suffix}": auth_nonce,
            f"x-auth-cert{auth_suffix}": b64encoded_der(self.certificate),
        }

    def log(self: Self, level: int, message: str, *args: ..., **kwargs: ...) -> None:
        """Log a message with the Apple ID as a prefix."""
        logger.log(level, f"[{self.apple_id}] " + f"{message}", *args, **kwargs, stacklevel=2)

    async def _send_authentication_request(self: Self, password: str, two_factor_code: str = "") -> None:
        """Send an authentication request to the IDS server."""
        if two_factor_code:
            self.log(logging.WARNING, "Two-factor authentication through IDS is not recommended. Please use GSA.")

        data = {
            "username": self.apple_id,
            "password": f"{password.strip()}{two_factor_code.strip()}",
        }

        self.log(logging.DEBUG, f"Authenticating through IDS via <{AUTHENTICATE_USER_URL}>.")

        response = await self._client.post(AUTHENTICATE_USER_URL, content=plistlib.dumps(data))
        response_data = plistlib.loads(response.content)
        status_code = AppleStatusCode(response_data.get("status"))

        self.log(logging.DEBUG, f"Response Payload: {json.dumps(response_data, indent=4)}")

        match status_code:
            case AppleStatusCode.SUCCESS:
                self.log(logging.DEBUG, "Successfully authenticated.")

            case AppleStatusCode.UNAUTHENTICATED:
                self.log(logging.WARNING, "User requires two-factor authentication.")
                raise IDSTwoFactorRequiredError

            case AppleStatusCode.ACTION_AUTHENTICATION_FAILED:
                if two_factor_code:
                    self.log(logging.ERROR, "Invalid two-factor authentication code.")
                    raise IDSTwoFactorInvalidError

            case AppleStatusCode.MISSING_REQUIRED_KEY:
                self.log(
                    logging.ERROR,
                    "2FA cannot be used to authenticate this account. Please use an app-specific password.",
                )

            case AppleStatusCode.ACTION_RETRY_WITH_DELAY:
                self._auth_backoff.max_delay = response_data.get("retry-interval", 60)

                sleep_time = self._auth_backoff.next()
                self.log(logging.INFO, f"Retrying authentication in {sleep_time} seconds.")
                await asyncio.sleep(sleep_time)

                await self._send_authentication_request(password, two_factor_code)
                return

            case _:
                self.log(logging.ERROR, f"Failed to authenticate. {status_code.name} ({status_code.value})")
                msg = f"Failed to authenticate. {status_code.name} ({status_code.value})"
                raise IDSError(msg)

        self.auth_token = response_data["auth-token"]
        self.profile_id = response_data["profile-id"]
        self._ensure_random_credentials()

    async def start_authentication(self: Self, password: str) -> None:
        """Start the authentication process."""
        self.password = password
        await self._gsa_client.authenticate(self.apple_id, password)

        match self._gsa_client.au_status:
            case AUStatus.COMPLETE:
                if self._gsa_client.pet_token is None:
                    msg = "GSA did not return a PET token."
                    raise ValueError(msg)

                self.log(logging.DEBUG, "Successfully authenticated through GSA.")
                await self._send_authentication_request(self._gsa_client.pet_token)

            case AUStatus.TRUSTED_DEVICE:
                await self._gsa_client.request_device_2fa()
                self.requires_2fa = True

            case AUStatus.SMS:
                msg = "SMS-based 2FA is not yet supported."
                raise NotImplementedError(msg)

    async def verify_2fa_code(self: Self, code: str) -> None:
        """Verify the two-factor authentication code."""
        success = await self._gsa_client.validate_device_2fa(code)
        await self._gsa_client.authenticate(self.apple_id, self.password)

        if success is True:
            await self._send_authentication_request(self._gsa_client.pet_token)
        else:
            await self._gsa_client.request_device_2fa()

        self.requires_2fa = not success

    async def authenticate_device(self: Self) -> None:
        """Authenticate the device with the IDS server."""
        if not self.private_key:
            self.private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

        csr = CertificateSigningRequestBuilder(
            subject_name=Name(
                [
                    NameAttribute(NameOID.COMMON_NAME, randbytes(20).hex()),
                ],
            ),
        ).sign(self.private_key, SHA256())

        payload = plistlib.dumps(
            {
                "authentication-data": {"auth-token": self.auth_token},
                "csr": b64decode(b64encoded_der(csr)),
                "realm-user-id": self.profile_id,
            },
        )

        response = await self._client.post(AUTHENTICATE_DEVICE_URL, content=payload)
        response_data = plistlib.loads(response.content)

        status_code = AppleStatusCode(response_data.get("status"))
        if status_code is not AppleStatusCode.SUCCESS:
            self.log(logging.ERROR, f"Failed to authenticate device. Status: {status_code.name}")
            msg = f"Failed to authenticate device. Status: {status_code.name}"
            raise IDSError(msg)

        self.certificate = load_der_x509_certificate(response_data["cert"])
        self.log(logging.DEBUG, "Successfully authenticated device.")
        self.log(
            logging.DEBUG,
            f"Device certificate is valid until {self.certificate.not_valid_after_utc.astimezone()}.",
        )

        self.on_device_authenticated_event.set()

        await self.fetch_handles()

    async def fetch_handles(self: Self) -> None:
        """Fetch the user's handles, verifying that the user is authenticated."""
        if self.is_device_authenticated is False:
            msg = "User is not authenticated. Device authentication is required to fetch handles."
            raise ValueError(msg)

        if self.push_token_provider is None:
            msg = "Push token is required to fetch handles."
            raise ValueError(msg)

        logger.debug(f"Fetching handles from {GET_HANDLES_URL}.")

        response = await self._client.get(
            GET_HANDLES_URL,
            headers=self._generate_auth_headers(GET_HANDLES_KEY),
        )

        response_data = plistlib.loads(response.content)

        status_code = AppleStatusCode(response_data.get("status"))
        match status_code:
            case AppleStatusCode.SUCCESS:
                self.handles = response_data["handles"]
                self.log(logging.DEBUG, f"Successfully fetched handles: {self.handles}")

            case _:
                self.log(logging.ERROR, f"Failed to fetch handles. Status: {status_code.name}")
                msg = f"Failed to fetch handles. Status: {status_code.name}"
                raise IDSError(msg)

    async def register_device(
        self: Self,
        device_name: str,
        validation_data: str | bytes,
        operating_system: str,
        product_type: str,
        os_version: str,
        build_version: str,
    ) -> None:
        """Register the device under the user."""
        if self.is_fully_authenticated is False:
            msg = "User is not fully authenticated. Device registration is not possible."
            raise ValueError(msg)

        if operating_system == "iOS":
            self.log(logging.WARNING, "iOS registration is not yet supported! Attempting registration despite this...")

        if isinstance(validation_data, str):
            validation_data = b64decode(validation_data)

        payload_data = {
            "language": "en-US",
            "device-name": device_name,
            "hardware-version": product_type,
            "os-version": f"{operating_system},{product_type},{build_version}",
            "software-version": build_version,
            "private-device-data": {
                "u": uuid4().hex.upper(),
            },
            "services": [
                {
                    "capabilities": [{"flags": 1, "name": "Messenger", "version": 1}],
                    "service": "com.apple.madrid",
                    "sub-services": [
                        "com.apple.private.alloy.sms",
                        "com.apple.private.alloy.gelato",
                        "com.apple.private.alloy.biz",
                        "com.apple.private.alloy.gamecenter.imessage",
                    ],
                    "users": [
                        {
                            "client-data": {
                                "is-c2k-equipment": True,
                                "optionally-receive-typing-indicators": True,
                                "public-message-identity-key": self.identity_key,
                                "public-message-identity-version": 2,
                                "show-peer-errors": True,
                                "supports-ack-v1": True,
                                "supports-activity-sharing-v1": True,
                                "supports-audio-messaging-v2": True,
                                "supports-autoloopvideo-v1": True,
                                "supports-be-v1": True,
                                "supports-ca-v1": True,
                                "supports-fsm-v1": True,
                                "supports-fsm-v2": True,
                                "supports-fsm-v3": True,
                                "supports-ii-v1": True,
                                "supports-impact-v1": True,
                                "supports-inline-attachments": True,
                                "supports-keep-receipts": True,
                                "supports-location-sharing": True,
                                "supports-media-v2": True,
                                "supports-photos-extension-v1": True,
                                "supports-st-v1": True,
                                "supports-update-attachments-v1": True,
                            },
                            "uris": self.handles,
                            "user-id": self.profile_id,
                        },
                    ],
                },
            ],
            "validation-data": validation_data,
        }
        payload = plistlib.dumps(payload_data)

        user_agent = f"com.apple.invitation-registration [Mac OS X,{os_version},{build_version},{product_type}]"

        response = await self._client.post(
            REGISTER_DEVICE_URL,
            content=payload,
            headers={
                "User-Agent": user_agent,
                "x-auth-user-id-0": self.profile_id,
                **self._generate_auth_headers(REGISTER_DEVICE_KEY, payload, auth_suffix_number=0),
            },
        )

        response_data = plistlib.loads(response.content)
        status_code = AppleStatusCode(response_data.get("status"))

        match status_code:
            case AppleStatusCode.SUCCESS:
                self.log(logging.DEBUG, "Successfully registered device.")

            case _:
                msg = f"Failed to register device. Status: {status_code.name} ({status_code.value})"
                raise IDSError(msg)

        try:
            certificate_data = response_data["services"][0]["users"][0]["cert"]

        except KeyError as e:
            msg = "Failed to obtain certificate data."
            raise IDSError(msg) from e

        self.device_certificate = load_der_x509_certificate(certificate_data)
        self.on_device_registered_event.set()

        self.log(logging.DEBUG, "Successfully obtained device certificate.")
        self.log(logging.DEBUG, f"Certificate valid until {self.device_certificate.not_valid_after_utc.astimezone()}.")

    def write_credentials(self: Self, path: PathLike) -> None:
        """Write the user's credentials to a file."""
        path = path if isinstance(path, Path) else Path(path)
        user_path = path / self.apple_id
        user_path.mkdir(parents=True, exist_ok=True)

        for name, file_name in self._credential_file_names.items():
            credential_path = user_path / file_name

            credential = getattr(self, name, None)
            if credential is None:
                continue

            match name:
                case "auth_token" | "profile_id":
                    credential_path.write_text(credential)

                case "private_key" | "signing_key" | "encryption_key":
                    key_data = credential.private_bytes(Encoding.PEM, PrivateFormat.TraditionalOpenSSL, NoEncryption())
                    credential_path.write_bytes(key_data)

                case "certificate" | "device_certificate":
                    certificate_data = credential.public_bytes(Encoding.PEM)
                    credential_path.write_bytes(certificate_data)

                case "identity_key":
                    credential_path.write_bytes(credential)

    def read_credentials(self: Self, path: PathLike) -> bool:
        """Read the user's credentials from a file."""
        path = path if isinstance(path, Path) else Path(path)

        if path.name != self.apple_id:
            path /= self.apple_id

        if not path.is_dir():
            return False

        self.log(logging.INFO, f"Reading credentials from {path}.")

        unknown = False
        for name, file_name in self._credential_file_names.items():
            credential_path = path / file_name
            if not credential_path.is_file():
                continue

            match name:
                case "auth_token" | "profile_id":
                    credential = credential_path.read_text()

                case "private_key" | "signing_key" | "encryption_key":
                    key_data = credential_path.read_bytes()
                    credential = load_pem_private_key(key_data, None)

                case "certificate" | "device_certificate":
                    certificate_data = credential_path.read_bytes()
                    credential = load_pem_x509_certificate(certificate_data)

                case "identity_key":
                    credential = credential_path.read_bytes()

                case _:
                    logger.debug(f"Unknown credential: {name}")
                    unknown = True
                    continue

            setattr(self, name, credential)

        self._ensure_random_credentials()

        return not unknown



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""Package for device-related classes and enums."""

from ._device import Device, OperatingSystem

__all__ = [
    "Device",
    "OperatingSystem",
]



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""APNs event listener for handling APNs commands and topic-specific push notifications."""

from __future__ import annotations

from hashlib import sha1
from typing import TYPE_CHECKING, Callable, Self

from .._util.event_listener import EventListener, EventType

if TYPE_CHECKING:
    from ..apns.protocol.packet import APNsCommand


class APNsListener(EventListener):
    """An event listener for APNs commands and topic-specific push notifications."""

    def __init__(self: Self) -> None:
        """Initialize the APNs listener."""
        super().__init__()
        self.enabled_topics: list[bytes] = []
        """A list of enabled topic hashes that will eventually be used to filter incoming push notifications."""

    def _register_command_listener(
        self: Self,
        command: type[APNsCommand],
        *,
        internal: bool = True,
    ) -> Callable[[APNsCommand], None]:
        """Register an internal listener for a specific APNs command."""
        return self._register_event_listener(EventType.COMMAND, command, internal=internal)

    def register_command_listener(self: Self, command: type[APNsCommand]) -> Callable[[APNsCommand], None]:
        """Register a listener for a specific APNs command."""
        return self._register_command_listener(command, internal=False)

    def _register_topic_listener(self: Self, topic: str | bytes, *, internal: bool = True) -> Callable:
        """Register an internal listener for a specific topic."""
        topic_hash = sha1(topic.encode(), usedforsecurity=False).digest() if isinstance(topic, str) else topic
        self.enabled_topics.append(topic_hash)
        return self._register_event_listener(EventType.TOPIC, topic_hash, internal=internal)

    def register_topic_listener(self: Self, topic: str | bytes) -> Callable:
        """Register a listener for a specific topic."""
        return self._register_topic_listener(topic, internal=False)

    def _register_connect_listener(self: Self, *, internal: bool = True) -> Callable:
        """Register an internal listener for the APNs connection event."""
        return self._register_event_listener(EventType.CONNECT, EventType.CONNECT.name, internal=internal)

    def register_connect_listener(self: Self) -> Callable:
        """Register a listener for the APNs connection event."""
        return self._register_connect_listener(internal=False)



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""APNs sub-package responsible for packet serialization and deserialization."""



#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel

from __future__ import annotations

import re
from base64 import b64encode
from random import SystemRandom
from typing import TYPE_CHECKING, Final, Protocol

from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    PublicFormat,
)

if TYPE_CHECKING:
    from cryptography.hazmat.primitives.asymmetric import ec, rsa
    from cryptography.hazmat.primitives.serialization import KeySerializationEncryption

PEM_EXPRESSION = re.compile(r"""-----(BEGIN|END) ([A-Z]+ ?)+-----""")

SYSTEM_RANDOM: Final = SystemRandom()
randint: Final = SYSTEM_RANDOM.randint
randbytes: Final = SYSTEM_RANDOM.randbytes
getrandbits = SYSTEM_RANDOM.getrandbits


class PrivateKey(Protocol):
    def private_bytes(
        self: PrivateKey,
        encoding: Encoding,
        format: PrivateFormat,  # noqa: A002
        encryption_algorithm: KeySerializationEncryption,
    ) -> bytes: ...


class PublicKey(Protocol):
    def public_bytes(
        self: PublicKey,
        encoding: Encoding,
    ) -> bytes: ...


def b64encoded_der(
    credential: PrivateKey | PublicKey | bytes | str,
) -> bytes:
    """
    Convert a credential to a base64-encoded DER.

    :param credential: A `cryptography` credential, or a credential (byte-)string in PEM format.
    :return: The base64-encoded DER representation of the credential.

    >>> from cryptography.hazmat.primitives.asymmetric import rsa
    >>> key = rsa.generate_private_key(65537, 2048)
    >>> expected = b64encoded_der(key.private_bytes(Encoding.PEM, PrivateFormat.TraditionalOpenSSL, NoEncryption()))
    >>> b64encoded_der(key) == expected
    True
    """
    if hasattr(credential, "private_bytes"):
        result = b64encode(credential.private_bytes(Encoding.DER, PrivateFormat.TraditionalOpenSSL, NoEncryption()))

    elif hasattr(credential, "public_bytes"):
        result = b64encode(credential.public_bytes(Encoding.DER))

    elif isinstance(credential, bytes):
        result = PEM_EXPRESSION.sub("", credential.decode()).strip().encode()

    elif isinstance(credential, str):
        result = PEM_EXPRESSION.sub("", credential).strip().encode()

    else:
        msg = f"Expected supported PEM, got {type(credential)}."
        raise TypeError(msg)

    return result.replace(b"\n", b"")


def construct_identity_key(
    public_signing_key: ec.EllipticCurvePublicKey,
    public_encryption_key: rsa.RSAPublicKey,
) -> bytes:
    """Construct an identity key from the public signing and encryption keys."""
    return (
        b"\x30\x81\xf6\x81\x43\x00\x41"
        + public_signing_key.public_bytes(Encoding.X962, PublicFormat.UncompressedPoint)
        + b"\x82\x81\xae\x00\xac"
        + public_encryption_key.public_bytes(Encoding.DER, PublicFormat.PKCS1)
    )
